package Generate::Main;

# Module:  Generate::Main
# Purpose: Generates a DTD from xml files
# Author:  John Storrs, Laboratory for Micro Enterprise
# License: FSF GPL
# Version: $Id: Main.pm,v 1.2 2005/07/10 11:37:07 lme Exp $

# The xml input files must conform to the DTD you want to capture, starting 
# from the same root.


use strict;

BEGIN {
   require Exporter;
   our @ISA = qw(Exporter);
   our @EXPORT_OK = qw(generateDtd %hasElements %hasAttributes);
}

use LME::XmlParser;
use LME::XmlNode;
use Generate::Element qw(generateElement);
use Generate::Attlist qw(generateAttlist);

# prototypes
sub generateDtd();
sub buildTagList($$);

# global variables
our (%hasElements, %hasAttributes);


sub generateDtd ()
{
   my ($infile, $outfile, $result, @files, $root, $rootTag, %tags, @tags,
       $tag, @nodes, $node, @children, $child, $childTag, $priorTag,
       @tagCounts, $i, $elementContent, $elementDefinition, $hasAttributes, 
       @roots, @attributes, $attribute, $attributeName, $attributeContent, 
       $attributeDefinition, @attributeContent);

   # grow an in-memory forest of xml trees
   @roots = ();
   @files = @ARGV;
   foreach $infile (@files) {
      $root = parseXml(importFile("$infile"));
      push(@roots, $root);
   }

   # check that the root tag is the same for all trees
   $rootTag = '';
   foreach $root (@roots) {
      if (!$rootTag) {
         $rootTag = $root->tag();
      } else {
         if ($root->tag() ne $rootTag) {
            print("ERROR: files have different root tags\n");
            exit(1);
         }
      }
   }

   # make an ordered list of all tags used
   %tags = ();
   foreach $root (@roots) {
      buildTagList($root, \%tags);
   }
   delete $tags{':cdata:'};
   delete $tags{':text:'};
   @tags = sort(keys(%tags));

   # extract element content model
   foreach $tag (@tags) {
      %hasElements = ();
      %hasAttributes = ();
      foreach $root (@roots) {
         # build a complex content model string
         @nodes = $root->nodeList($tag);
         if ($tag eq $rootTag) {
            push(@nodes, $root);
         }
         foreach $node (@nodes) {
            # record node attribute content
            @attributes = $node->attributeList();
            $attributeContent = '';
            foreach $attribute (@attributes) {
               $attributeName = $attribute->[0];
               $attributeContent .= "$attributeName, ";
            }
            if ($attributeContent) {
               $attributeContent =~ s/, $//;  # strip trailing ', '
               $hasAttributes{$attributeContent} += 1;
            }

            # record node element content
            $elementContent = '';
            @children = $node->children();
            if (scalar(@children) == 0) {
               $elementContent .= "empty, ";
            }  # ELSE PREFERRED HERE
            $priorTag = '';
            $i = 0;
            # look for a sequence of the same element
            foreach $child (@children) {
               $childTag = $child->tag();
               if ($childTag eq $priorTag) {
                  $i++;
               } else {
                  if ($priorTag ne '') {
                     if ($i > 1) {
                        $elementContent .= "$priorTag+, ";
                     } else {
                        $elementContent .= "$priorTag, ";
                     }
                  }
                  $i = 1;
               }
               $priorTag = $childTag;
            }
            if ($priorTag ne '') {
               if ($i > 1) {
                  $elementContent .= "$priorTag+, ";
               } else {
                  $elementContent .= "$priorTag, ";
               }
            }
            $hasElements{$elementContent} = 1;
         }
      }

      $result .= generateElement($tag);
      $result .= generateAttlist($tag);
   }

   # save result
   $outfile = "$rootTag.dtd";
   open(OUTFILE, ">$outfile") or die("can't open $outfile\n");
   print(OUTFILE $result);
   close(OUTFILE) or die("can't close $outfile\n");
}


sub buildTagList ($$)
{
   my ($node, $tags) = @_;
   my ($tag, $child, @children);

   # record tag in %tags hash
   $tag = $node->tag();
   $$tags{$tag} = 1;

   # recurse
   @children = $node->children();
   foreach $child (@children) {
      buildTagList($child, $tags);
   }
}


return 1;
