package Journal::Main;

# Module:  Journal::Main
# Purpose: Generates documentation from journal xml files.
# Author:  John Storrs, Laboratory for Micro Enterprise
# License: FSF GPL
# Version: $Id: Main.pm,v 1.2 2005/07/10 11:37:07 lme Exp $

# HANDLE EMPTY TOPICS BETTER. RETURN A 'HAS CONTENT' FLAG RATHER THAN USING
# A CRUDE <para>\s*[a-z..... TEST.


use strict;

BEGIN {
   require Exporter;
   our @ISA = qw(Exporter);
   our @EXPORT = qw(processJournal);
}

use LME::XmlParser;
use LME::XmlNode;
use Text::Wrap;


# forward declarations
sub processJournal($);
sub processByTopic($$);


# global variables
our (%topicUse);


# sub processJournal:
# Generate DocBook documentation from journal entries, in date and in topic 
# order. Entries with topic attributes not found in the topic tree are ignored,
# and the topic is listed as unrecognised. The topicUse hash maps key(topic) 
# with value(date ordered list of entries), for fast processing. 

sub processJournal ($)
{
   my ($topicHome) = @_;
   my ($result, $journal, @entries, $entry, $topicNames, $topicName, $topicId,
       @topicNames, $filename, $topic, $rootTopic, @allTopics, @topicEntries, 
       @topics, %isTopic, $line, $contents); 

   %topicUse = ();

   # remove old generated files
   system("rm -f ../docbook/*");
   system("rm -f ../pdf/*");

   $filename = "$topicHome/xml/main.xml";
   $rootTopic = parseXml(importFile($filename));
   @allTopics = $rootTopic->nodeList('topic');
   push(@allTopics, $rootTopic);
   foreach $topic (@allTopics) {
      $topicName = $topic->attribute('name');
      $topicId = $topic->attribute('id');
      $isTopic{"$topicName ($topicId)"} = 1;
   }

   $filename = 'journal.xml';
   $journal = parseXml(importFile($filename));
   @entries = $journal->nodeList('entry');

   # build topic use hash (here topic names have the form 'name (id)')
   foreach $entry (@entries) {
      $topicNames = $entry->attribute('topic');
      @topicNames = split(', ', $topicNames);
      foreach $topicName (@topicNames) {
         $topicName =~ m/\(([^)]+)\)/;
         $topicId = $1;
         push(@{$topicUse{$topicId}}, $entry);
         if (!$isTopic{$topicName}) {
            print("$topicName is not a recognised topic\n");
         }
      }
   }

   # make a book for the journal, organised by date and by topic

   $result = <<EOT;
<book>
  <title>Journal</title>

  <para><ulink url="../pdf/journal.pdf">pdf</ulink></para>

  <section class="topic" id="bydate">
    <title>By Date</title>

EOT

   $result .= processByDate($journal);

   $result .= <<EOT;

  </section>

EOT

   $result .= <<EOT;

  <section class="topic" id="byTopic">
    <title>By Topic</title>

EOT

   # recursively process topics
   $result .= processByTopic($rootTopic, 1);

   $result .= <<EOT;

  </section>

</book>
EOT

   $filename = '../docbook/journal.xml';
   open(FILE, ">$filename") or die "can't open $filename";
   print(FILE $result);
   close(FILE) or die "can't close $filename";
   system("$ENV{PXML_HOME}/app/publishxml/publishxml html.cfg $filename");
   system("$ENV{PXML_HOME}/app/publishxml/publishxml pdf.cfg $filename");

   # make separate articles for each topic in the journal
   foreach $topic (@allTopics) {
      $topicName = $topic->attribute('name');
      $topicId = $topic->attribute('id');

      $result = <<EOT;
<article>
  <title>$topicName ($topicId)</title>

EOT

      # recursively process topics
      $result .= processByTopic($topic, 1);
      if ($result !~ m/<para>/) {
         next;  # no content, so skip
      }

      $result .= <<EOT;

   </article>
EOT

      $topicName =~ s/\s//g;  # strip spaces
      $topicName =~ s/\///g;  # strip '/'
      $filename = "../docbook/${topicName}-$topicId.xml";
      open(FILE, ">$filename") or die "can't open $filename";
      print(FILE $result);
      close(FILE) or die "can't close $filename";
      system("$ENV{PXML_HOME}/app/publishxml/publishxml pdf.cfg $filename");
  }
}


sub processByDate ($)
{
   my ($journal) = @_;
   my ($entry, $entryTopic, $child, $entryDate, $result, $dayName, $dayNumber, 
       $monthNumber, $yearNumber, $currentYear, $currentMonth, $currentDay,
       $started, $shortYear, @entries, @children, @dates, %monthName, 
       %entriesByDate);

   @dates = ();
   %entriesByDate = ();
   $currentYear = 0;
   $currentMonth = 0;
   $currentDay = 0;
   $started = 0;
   %monthName = (1 => 'January', 2 => 'February', 3 => 'March', 4 => 'April',
                 5 => 'May', 6 => 'June', 7 => 'July', 8 => 'August',
                 9 => 'September', 10 => 'October', 11 => 'November', 
                 12 => 'December', 0 => 'ERROR');

   @entries = $journal->nodeList('entry');
   foreach $entry (@entries) {
      $entryTopic = $entry->attribute('topic');
      $entryDate = $entry->attribute('date');

      # extract date components from format 'Mon 27/5/02'
      $entryDate =~ m|(\w+)\s(\d+)/\s*(\d+)/\s*(\d+)|;
      $dayName = $1;
      $dayNumber = $2;
      $monthNumber = $3;
      $yearNumber = $4;
      if (!$dayNumber) {
         $dayNumber = 0;
      }
      if (!$monthNumber) {
         $monthNumber = 0;
      }
      if (!$yearNumber) {
         $yearNumber = 0;
      }
      $shortYear = $yearNumber;
      if ($yearNumber > 50) {
         $yearNumber += 1900; 
      } else {
         $yearNumber += 2000; 
      }

      if ($yearNumber != $currentYear) {
         $currentYear = $yearNumber;
         $currentMonth = $monthNumber;
         $currentDay = $dayNumber;

         if ($started) {
            $result .= <<EOT;
        </section>
      </section>
    </section>
EOT

         }

         $result .= <<EOT;
    <section class="topic" id="$currentYear">
      <title>$currentYear</title>
      <section class="topic" id="$currentMonth$currentYear">
        <title>$monthName{$currentMonth}</title>
        <section>
          <title>$dayNumber/$monthNumber/$shortYear</title>
EOT

      }

      if ($monthNumber != $currentMonth) {
         $currentMonth = $monthNumber;
         $currentDay = $dayNumber;
         if ($started) {
            $result .= <<EOT;
        </section>
      </section>
EOT

         }

         $result .= <<EOT;
      <section class="topic" id="$currentMonth$currentYear">
        <title>$monthName{$currentMonth}</title>
        <section>
          <title>$dayNumber/$monthNumber/$shortYear</title>
EOT

      }

      if ($dayNumber != $currentDay) {
         $currentDay = $dayNumber;
         if ($started) {
            $result .= <<EOT;
        </section>
EOT

         }
         $result .= <<EOT;
        <section>
          <title>$dayNumber/$monthNumber/$shortYear</title>
EOT

      }

      $result .= <<EOT;
          <para><command>$entryTopic</command></para>
EOT

      @children = $entry->children();
      foreach $child (@children) {
         $result .= $child->generateXml();
      }

      $started = 1;
   }


   $result .= <<EOT;
        </section>
      </section>
    </section>
EOT

   return $result;
}

      
# sub processByTopic

sub processByTopic ($$)
{
   my ($topic, $starting) = @_;
   my ($topicName, $topicId, $result, $entry, $date, $child, $subTopic, 
       $content, @topicEntries, @children, @subTopics, );

   $topicName = $topic->attribute('name');
   $topicId = $topic->attribute('id');
   if ($topicUse{$topicId}) {
      @topicEntries = @{$topicUse{$topicId}};
   } else {
      @topicEntries = ();
   }

   if (!$starting) {

      $result = <<EOT;
  <section class="topic" id="topic$topicId">
    <title>$topicName ($topicId)</title>

EOT

   }

   $topicName =~ s/\s//g;  # strip spaces
   $topicName =~ s/\///g;  # strip '/'

   $result .= <<EOT;
    <para><ulink url="../pdf/${topicName}-$topicId.pdf">pdf</ulink></para>

EOT

   foreach $entry (@topicEntries) {
      $date = $entry->attribute('date');
      @children = $entry->children();

      $content = '';
      foreach $child (@children) {
         $content .= $child->generateXml();
      }

      $result .= <<EOT;
      <para><command>$date</command></para>
$content
EOT

   }

   # recursively process sub-topics
   @subTopics = $topic->nodeList('topic', 1);
   foreach $subTopic (@subTopics) {
      $result .= processByTopic($subTopic, 0);
   }

   if (!$starting) {

      $result .= <<EOT;
  </section>

EOT

   }

   if ($result =~ m/<para>\s*[a-zA-Z0-9]/) {  # subtree contains entries
      return $result;
   } else {  # throw topic headings away
      return '';  
   }
}


return 1;
