package Records::Main;

# Module:  Records::Main
# Purpose: Generates documentation from record xml files.
# Author:  John Storrs, Laboratory for Micro Enterprise
# License: FSF GPL
# Version: $Id: Main.pm,v 1.2 2004/06/05 17:51:04 lme Exp $


use strict;

BEGIN {
   require Exporter;
   our @ISA = qw(Exporter);
   our @EXPORT = qw(processRecords);
}

use LME::XmlParser;
use LME::XmlNode;


# forward declarations
sub processRecords($);
sub processByTopic($$);


# global variables
our (%topicUse);


# sub processRecords:
# Generates DocBook documentation from project records, in standard topic tree
# order. Records with topic attributes not found in the standard topic tree are 
# identified. The topicUse hash maps key(topic id) with value(list of records),
# for fast processing. 

sub processRecords ($)
{ 
   my ($topicHome) = @_;
   my ($result, $records, @records, $record, $topicNames, $topicName, $topicId,
       @topicNames, $filename, $rootTopic, $rootTopicName, $topic, 
       @topicRecords, @allTopics, %isTopic, $contents, @contents);

   %topicUse = ();

   # remove old generated files
   system("rm -f ../docbook/*");
   system("rm -f ../pdf/*");

   $filename = "$topicHome/xml/main.xml";
   $rootTopic = parseXml(importFile($filename));
   $rootTopicName = $rootTopic->attribute('name');
   @allTopics = $rootTopic->nodeList('topic');
   push(@allTopics, $rootTopic);
   foreach $topic (@allTopics) {
      $topicName = $topic->attribute('name');
      $topicId = $topic->attribute('id');
      $isTopic{"$topicName ($topicId)"} = 1;
   }

   # build include file
   $filename = 'records.xml';
   open(FILE, ">$filename") or die "can't open $filename";
   print(FILE "<!-- GENERATED FILE - DO NOT EDIT -->\n");
   print(FILE "<records>\n");
   foreach $topic (@allTopics) {
      $topicName = $topic->attribute('name');
      $topicId = $topic->attribute('id');
      $topicName =~ s/\s//g;  # strip spaces
      $topicName =~ s/\///g;  # strip '/'
      if (-f "${topicName}-$topicId.xml") {
         print(FILE "  <include file=\"${topicName}-$topicId.xml\"/>\n");
      }
   }
   print(FILE "</records>\n");
   close(FILE) or die "can't close $filename";

   # build record node tree
   $records = parseXml(importFile($filename));
   @records = $records->nodeList('record');

   # build topic use hash (here topic names have the form 'name (id)')
   foreach $record (@records) {
      $topicNames = $record->attribute('topic');
      @topicNames = split(', ', $topicNames);
      foreach $topicName (@topicNames) {
         $topicName =~ m/\(([^)]+)\)/;
         $topicId = $1;
         push(@{$topicUse{$topicId}}, $record);
         if (!$isTopic{$topicName}) {
            print("$topicName is not a recognised topic\n");
         }
      }
   }

   # make an article organised by topic

   $result = <<EOT;
<article>
  <title>$rootTopicName</title>

  <!--<para><ulink url="../pdf/records.pdf">pdf</ulink></para>-->

EOT

   # recursively process topics
   $result .= processByTopic($rootTopic, 1);

   $result .= <<EOT;
</article>
EOT

   $filename = '../docbook/records.xml';
   open(FILE, ">$filename") or die "can't open $filename";
   print(FILE $result);
   close(FILE) or die "can't close $filename";
   system("$ENV{PXML_HOME}/app/publishxml/publishxml html.cfg $filename");
   system("$ENV{PXML_HOME}/app/publishxml/publishxml pdf.cfg $filename");

   # make separate articles for each topic in the records
   foreach $topic (@allTopics) {
      $topicName = $topic->attribute('name');
      $topicId = $topic->attribute('id');

=hide
      $result = <<EOT;
<article>
  <title>$topicName ($topicId)</title>

EOT
=cut

      $result = <<EOT;
<article>
  <title>$topicName</title>

EOT

      # recursively process topics
      $result .= processByTopic($topic, 1);
      if ($result !~ m/<para>/) {
         next;  # no content, so skip
      }

      $result .= <<EOT;

   </article>
EOT

      $topicName =~ s/\s//g;  # strip spaces
      $topicName =~ s/\///g;  # strip '/'
      $filename = "../docbook/${topicName}-$topicId.xml";
      open(FILE, ">$filename") or die "can't open $filename";
      print(FILE $result);
      close(FILE) or die "can't close $filename";
      system("$ENV{PXML_HOME}/app/publishxml/publishxml pdf.cfg $filename");
  }
}

      
sub processByTopic ($$)
{
   my ($topic, $starting) = @_;
   my ($topicName, $topicId, $fileName, $result, $record, $child, $subTopic, 
       @topicRecords, @children, @subTopics);

   $topicName = $topic->attribute('name');
   $topicId = $topic->attribute('id');
   if ($topicUse{$topicId}) {
      @topicRecords = @{$topicUse{$topicId}};
   } else {
      @topicRecords = ();
   }

   $fileName = "$topicName-$topicId";
   $fileName =~ s/ //g;  # strip spaces
   $fileName =~ s/\///g;  # strip '/'

   if (!$starting) {

=hide
      $result = <<EOT;
  <section id="$fileName">
    <title>$topicName ($topicId)</title>

EOT
=cut

      $result = <<EOT;
  <section class="topic" id="$fileName">
    <title>$topicName</title>

EOT

   }

   $result .= <<EOT;
    <para><ulink url="../pdf/$fileName.pdf">pdf</ulink></para>

EOT

   foreach $record (@topicRecords) {
      @children = $record->children();
      foreach $child (@children) {
         $result .= $child->generateXml();
      }
   }

   # recursively process sub-topics
   @subTopics = $topic->nodeList('topic', 1);  # children only
   foreach $subTopic (@subTopics) {
      $result .= processByTopic($subTopic, 0);
   }

   if (!$starting) {

      $result .= <<EOT;
  </section>

EOT

   }

   if ($result =~ m/<para>[a-zA-Z0-9]/) {  # subtree contains records
      return $result;
   } else {  # throw topic headings away
      return '';  
   }
}


return 1;
