package Tasks::Main;

# Module:  Tasks::Main
# Purpose: Generates documentation from task xml files.
# Author:  John Storrs, Laboratory for Micro Enterprise
# License: FSF GPL
# Version: $Id: Main.pm,v 1.2 2005/07/10 11:37:07 lme Exp $


use strict;

BEGIN {
   require Exporter;
   our @ISA = qw(Exporter);
   our @EXPORT = qw(processTasks);
}

use LME::XmlParser;
use LME::XmlNode;


# forward declarations
sub processTasks($);
sub processByPriority($);
sub processByTopic($$);


# global variables
our (%topicUse);


# sub processTasks:
# Generates DocBook documentation for project tasks, in priority and in topic 
# tree order. Tasks with topic attributes not found in the topic tree are 
# identified. The topicUse hash maps key(topic id) with value(list of tasks),
# for fast processing. 

sub processTasks ($)
{ 
   my ($topicHome) = @_;
   my ($result, $tasks, @tasks, $task, $topicNames, $topicName, $topicId,
       @topicNames, $filename, $rootTopic, $topic, @topicTasks, @allTopics, 
       %isTopic, $contents, @contents);

   %topicUse = ();

   # remove old generated files
   system("rm -f ../docbook/*");
   system("rm -f ../pdf/*");

   $filename = "$topicHome/xml/main.xml";
   $rootTopic = parseXml(importFile($filename));
   @allTopics = $rootTopic->nodeList('topic');
   push(@allTopics, $rootTopic);
   foreach $topic (@allTopics) {
      $topicName = $topic->attribute('name');
      $topicId = $topic->attribute('id');
      $isTopic{"$topicName ($topicId)"} = 1;
   }

   # build task node tree
   $filename = 'tasks.xml';
   $tasks = parseXml(importFile($filename));
   @tasks = $tasks->nodeList('task');

   # build topic use hash (here topic names have the form 'name (id)')
   foreach $task (@tasks) {
      $topicNames = $task->attribute('topic');
      @topicNames = split(', ', $topicNames);
      foreach $topicName (@topicNames) {
         $topicName =~ m/\(([^)]+)\)/;
         $topicId = $1;
         push(@{$topicUse{$topicId}}, $task);
         if (!$isTopic{$topicName}) {
            print("$topicName is not a recognised topic\n");
         }
      }
   }

   # make an article for open tasks, organised by priority and by topic

   $result = <<EOT;
<article>
  <title>Tasks</title>

  <section class="topic" id="bypriority">>
    <title>Open Tasks by Priority</title>

    <para><ulink url="../pdf/tasks.pdf">pdf</ulink></para>

    <para>For now the pdf above covers everything. Soon there will be pdfs for 
      open tasks by priority, for each owner etc. Priority sorting is simple
      ascii. I suggest starting with a letter, then digits. Presentation could
      then also be in groups by letter representing urgency.</para>

EOT

   $result .= processByPriority($tasks);
  
   $result .= <<EOT;

  </section>

  <section class="topic" id="bytopic">
    <title>All Tasks by Topic</title>

EOT

   # recursively process topics
   $result .= processByTopic($rootTopic, 1);

   $result .= <<EOT;

  </section>

</article>
EOT

   $filename = '../docbook/tasks.xml';
   open(FILE, ">$filename") or die "can't open $filename";
   print(FILE $result);
   close(FILE) or die "can't close $filename";
   system("$ENV{PXML_HOME}/app/publishxml/publishxml html.cfg $filename");
   system("$ENV{PXML_HOME}/app/publishxml/publishxml pdf.cfg $filename");

   # make separate articles for each topic in the tasks
   foreach $topic (@allTopics) {
      $topicName = $topic->attribute('name');
      $topicId = $topic->attribute('id');

      $result = <<EOT;
<article>
  <title>$topicName ($topicId)</title>

EOT

      # recursively process topics
      $result .= processByTopic($topic, 1);
      if ($result !~ m/<para>/) {
         next;  # no content, so skip
      }

      $result .= <<EOT;

   </article>
EOT

      $topicName =~ s/\s//g;  # strip spaces
      $topicName =~ s/\///g;  # strip '/'
      $filename = "../docbook/${topicName}-$topicId.xml";
      open(FILE, ">$filename") or die "can't open $filename";
      print(FILE $result);
      close(FILE) or die "can't close $filename";
      system("$ENV{PXML_HOME}/app/publishxml/publishxml pdf.cfg $filename");
  }
}


# sub processByPriority:
# Sorts output first by owner, then by priority in groups. Priorities are 
# strings consisting of a capital letter followed by digits. Priority order is
# ascii sort order for the strings. The priority grouping is based on the 
# letter used. GROUP SORTING POSTPONED.

sub processByPriority ($)
{
   my ($tasks) = @_;
   my ($topicName, $topicId, $result, $task, $child, $subTopic, $taskId, 
       $taskTopic, $taskSummary, $taskOwner, $taskPriority, $taskStatus, 
       $taskDuration, $taskCreated, $taskRequired, $taskStart, $taskFinish, 
       $taskDetails, $taskAction, $owner, @tasks, @children, 
       @subTopics, %openTasks, @openTasks, %isTaskOwner, @taskOwners);

   # build sorting hashes
   %isTaskOwner = ();
   %openTasks = ();
   @tasks = $tasks->nodeList('task');
   foreach $task (@tasks) {
      $taskOwner = $task->attribute('owner');
      $taskPriority = $task->attribute('priority');
      $taskStatus = $task->attribute('status');
      $isTaskOwner{$taskOwner} = 1;
      if ($taskStatus ne '100%') {
         push(@{$openTasks{$taskPriority}}, $task);
      }
   }

   # build list of open tasks ordered by priority
   @openTasks = ();
   foreach $taskPriority (sort(keys(%openTasks))) {
      push(@openTasks, @{$openTasks{$taskPriority}});
   }

   # build list of task owners
   @taskOwners = ();
   foreach $taskOwner (sort(keys(%isTaskOwner))) {
      push(@taskOwners, $taskOwner);
   }

   foreach $owner (@taskOwners) {

      $result .= <<EOT;
    <section>
      <title>Task Owner <emphasis>$owner</emphasis></title>

EOT

      foreach $task (@openTasks) {
         $taskId = $task->attribute('id');
         $taskTopic = $task->attribute('topic');
         $taskSummary = $task->attribute('summary');
         $taskOwner  = $task->attribute('owner');
         $taskPriority = $task->attribute('priority');
         $taskStatus = $task->attribute('status');
         $taskDuration = $task->attribute('duration');
         $taskCreated = $task->attribute('created');
         $taskRequired = $task->attribute('required');
         $taskStart = $task->attribute('start');
         $taskFinish = $task->attribute('finish');

         if ($taskOwner ne $owner) {
            next;
         }
 
         $result .= <<EOT;
      <section>
        <title>$taskTopic: <emphasis>$taskSummary</emphasis></title>
  
        <informaltable>
          <tgroup cols="9">
            <colspec colwidth="18.8mm"/>
            <colspec colwidth="18.8mm"/>
            <colspec colwidth="18.8mm"/>
            <colspec colwidth="18.8mm"/>
            <colspec colwidth="18.8mm"/>
            <colspec colwidth="18.8mm"/>
            <colspec colwidth="18.8mm"/>
            <colspec colwidth="18.8mm"/>
            <colspec colwidth="18.8mm"/>
            <thead>
              <row>
                <entry>task id</entry>
                <entry>owner</entry>
                <entry>priority</entry>
                <entry>status</entry>
                <entry>duration</entry>
                <entry>created</entry>
                <entry>required</entry>
                <entry>start</entry>
                <entry>finish</entry>
              </row>
            </thead>
            <tbody>
              <row>
                <entry>$taskId</entry>
                <entry>$taskOwner</entry>
                <entry>$taskPriority</entry>
                <entry>$taskStatus</entry>
                <entry>$taskDuration</entry>
                <entry>$taskCreated</entry>
                <entry>$taskRequired</entry>
                <entry>$taskStart</entry>
                <entry>$taskFinish</entry>
              </row>
            </tbody>
          </tgroup>
        </informaltable>
      
EOT

         $taskDetails = $task->child('details');
         if ($taskDetails) {

            $result .= <<EOT;

        <para><command>Details</command></para>

EOT

            @children = $taskDetails->children();
            foreach $child (@children) {
               $result .= $child->generateXml();
            }
         }
   
         $taskAction = $task->child('action');
         if ($taskAction && $taskAction->children()) {

            $result .= <<EOT;

        <para><command>Action</command></para>

EOT

            @children = $taskAction->children();
            foreach $child (@children) {
               $result .= $child->generateXml();
            }
         }

         $result .= <<EOT;
      </section>

EOT

      }

      $result .= <<EOT;
    </section>

EOT

   }

   return $result;
}

      
sub processByTopic ($$)
{
   my ($topic, $starting) = @_;
   my ($topicName, $topicId, $result, $task, $child, $subTopic, $taskId, 
       $taskTopic, $taskSummary, $taskOwner, $taskPriority, $taskStatus, 
       $taskDuration, $taskCreated, $taskRequired, $taskStart, $taskFinish, 
       $taskDetails, $taskAction, @topicTasks, @children, @subTopics);

   $topicName = $topic->attribute('name');
   $topicId = $topic->attribute('id');
   if ($topicUse{$topicId}) {
      @topicTasks = @{$topicUse{$topicId}};
   } else {
      @topicTasks = ();
   }

   if (!$starting) {

      $result = <<EOT;
  <section>
    <title>$topicName ($topicId)</title>

EOT

   }

   $topicName =~ s/\s//g;  # strip spaces
   $topicName =~ s/\///g;  # strip '/'

   $result .= <<EOT;
    <para><ulink url="../pdf/${topicName}-$topicId.pdf">pdf</ulink></para>

EOT

   foreach $task (@topicTasks) {
      $taskId = $task->attribute('id');
      $taskTopic = $task->attribute('topic');
      $taskSummary = $task->attribute('summary');
      $taskOwner  = $task->attribute('owner');
      $taskPriority = $task->attribute('priority');
      $taskStatus = $task->attribute('status');
      $taskDuration = $task->attribute('duration');
      $taskCreated = $task->attribute('created');
      $taskRequired = $task->attribute('required');
      $taskStart = $task->attribute('start');
      $taskFinish = $task->attribute('finish');

      #if ($taskStatus eq '100%') {
      #   next;
      #}

      $result .= <<EOT;
    <section>
      <title><emphasis>$taskSummary</emphasis></title>

      <informaltable>
        <tgroup cols="9">
          <colspec colwidth="18.8mm"/>
          <colspec colwidth="18.8mm"/>
          <colspec colwidth="18.8mm"/>
          <colspec colwidth="18.8mm"/>
          <colspec colwidth="18.8mm"/>
          <colspec colwidth="18.8mm"/>
          <colspec colwidth="18.8mm"/>
          <colspec colwidth="18.8mm"/>
          <colspec colwidth="18.8mm"/>
          <thead>
            <row>
              <entry>task id</entry>
              <entry>owner</entry>
              <entry>priority</entry>
              <entry>status</entry>
              <entry>duration</entry>
              <entry>created</entry>
              <entry>required</entry>
              <entry>start</entry>
              <entry>finish</entry>
            </row>
          </thead>
          <tbody>
            <row>
              <entry>$taskId</entry>
              <entry>$taskOwner</entry>
              <entry>$taskPriority</entry>
              <entry>$taskStatus</entry>
              <entry>$taskDuration</entry>
              <entry>$taskCreated</entry>
              <entry>$taskRequired</entry>
              <entry>$taskStart</entry>
              <entry>$taskFinish</entry>
            </row>
          </tbody>
        </tgroup>
      </informaltable>
      
EOT

      $taskDetails = $task->child('details');
      if ($taskDetails) {

         $result .= <<EOT;

      <para><command>Details</command></para>

EOT

         @children = $taskDetails->children();
         foreach $child (@children) {
            $result .= $child->generateXml();
         }
      }

      $taskAction = $task->child('action');
      if ($taskAction && $taskAction->children()) {

         $result .= <<EOT;

      <para><command>Action</command></para>

EOT

         @children = $taskAction->children();
         foreach $child (@children) {
            $result .= $child->generateXml();
         }
      }

      $result .= <<EOT;
    </section>

EOT

   }

   # recursively process sub-topics
   @subTopics = $topic->nodeList('topic', 1);  # children only
   foreach $subTopic (@subTopics) {
      $result .= processByTopic($subTopic, 0);
   }

   if (!$starting) {

      $result .= <<EOT;
  </section>

EOT

   }

   if ($result =~ m/<para>[a-zA-Z0-9]/) {  # subtree contains tasks
      return $result;
   } else {  # throw topic headings away
      return '';  
   }
}


return 1;
