package Topics::Main;

# Module:  Topics::Main
# Purpose: Generates documentation from topic xml files.
# Author:  John Storrs, Laboratory for Micro Enterprise
# License: FSF GPL
# Version: $Id: Main.pm,v 1.2 2005/07/10 11:37:07 lme Exp $

use strict;
use LME::XmlParser;
use LME::XmlNode;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(processTopics);

# global variables
our (@topicById, @parentById, @depthById, $maxDepth, $result);

# forward declarations
sub buildParentByIdArray($$);
sub findMaxDepth($$);
sub processTopics();
sub generateTreeLabels($$);
sub generatePlainLabels($);
sub generateGraph($$);
sub generateDocBook($$);
sub makeSedFilter();
sub makeGroffMacros();    


sub processTopics ()
{
   my ($infile, $outFile, $rootTopic, $topic, $currentTopic, $currentPath, 
       $id, $parentId, $name, $order, $down, $inFile, $sedFilter, 
       $groffMacros, @topics);

   # get input file name
   $inFile = $ARGV[0];
   if (!$inFile) {
      die("usage: maketopics <input file>\n");
   }

   # check that topics file exists
   if (!(-e $inFile)) {
      die("can't find $inFile\n");
   }

   # build in-memory topic tree
   print("processing $inFile\n");
   $rootTopic = parseXml(importFile($inFile));

   # build @topicById just to check number is unique for now
   @topicById = ();
   @topics = $rootTopic->nodeList('topic');
   push(@topics, $rootTopic);
   foreach $topic (@topics) {
      $id = $topic->attribute('id');
      if ($topicById[$id]) {
         print("ERROR: duplicate topic id $id\n");
      } else {
         $topicById[$id] = $topic;
      }
   }

   # generate tree labels
   $result = '';
   generateTreeLabels($rootTopic, '');
   $outFile = "treelabels.txt";
   open(OUTFILE, ">$outFile") or die("can't open $outFile\n");
   print(OUTFILE $result);
   close(OUTFILE) or die("can't close $outFile\n");
   makeSedFilter();
   makeGroffMacros();
   system("sed -f filter $outFile >temp");
   system("groff macros temp > treelabels.ps 2> /dev/null");
   system("rm temp");
   system("ps2pdf treelabels.ps ../pdf/treelabels.pdf");

   # generate plain labels

   $result = <<EOT;
.fam P
.ps 12
.vs 55
.po -1.5c
EOT

   generatePlainLabels($rootTopic);

   $outFile = 'plainlabels.txt';
   open(OUTFILE, ">$outFile") or die("can't open $outFile\n");
   print(OUTFILE $result);
   close(OUTFILE) or die("can't close $outFile\n");
   system("groff $outFile > plainlabels.ps");
   system("ps2pdf plainlabels.ps ../pdf/plainlabels.pdf");

   # generate topic graph

   $result = <<EOT;
digraph topicgraph {
   graph [size="8,12", center=true, orientation=portrait, rankdir=LR, nodesep=0.2, ranksep=2];
   node [shape=record, fontsize=10, height=0.2, width=2.2];
   edge [style="setlinewidth(0.2)"];
EOT

   generateGraph($rootTopic, 0);

   $result .= <<EOT;
}
EOT

   $outFile = "topicgraph.dot";
   open(OUTFILE, ">$outFile") or die("can't open $outFile\n");
   print(OUTFILE $result);
   close(OUTFILE) or die("can't close $outFile\n");
   system("dot -Tps $outFile > topicgraph.ps");
   system("ps2pdf topicgraph.ps ../pdf/topicgraph.pdf");

   # generate and publish DocBook

   $result = <<EOT;
<article>
<title>Topics</title>

EOT

   generateDocBook($rootTopic, 0);

   $result .= <<EOT;

</article>
EOT

   $outFile = "../docbook/topics.xml";
   open(OUTFILE, ">$outFile") or die("can't open $outFile\n");
   print(OUTFILE $result);
   close(OUTFILE) or die("can't close $outFile\n");
   system("$ENV{PXML_HOME}/app/publishxml/publishxml html.cfg $outFile");
   system("$ENV{PXML_HOME}/app/publishxml/publishxml pdf.cfg $outFile");
}


# sub generateTreeLabels:
# Recursively generates tree labels for the topics.

sub generateTreeLabels ($$)
{
   my ($topic, $path) = @_;
   my ($name, $id, $subTopic, @subTopics);

   $name = $topic->attribute('name');
   $id = $topic->attribute('id');

   if ($name) {
      if ($path) {
         $path = "$path|$name";
      } else {
         $path = "$name";
      }
   } else {
      $path = '';
   }
   $result .= "$path ($id)\n";

   @subTopics = $topic->nodeList('topic', 1);  # children only
   foreach $subTopic (@subTopics) {
      generateTreeLabels($subTopic, $path);
   }
}


# sub generatePlainLabels:
# Recursively generates plain labels for the topics.

sub generatePlainLabels ($)
{
   my ($topic) = @_;
   my ($name, $id, $subTopic, @subTopics);

   $name = $topic->attribute('name');
   $id = $topic->attribute('id');

   $result .= "$name ($id)\n";
   $result .= ".br\n";

   @subTopics = $topic->nodeList('topic', 1);  # children only
   foreach $subTopic (@subTopics) {
      generatePlainLabels($subTopic);
   }
}


# sub generateGraph:
# Recursively generates a graph representation of the topic tree.

sub generateGraph ($$)
{
   my ($topic, $parentId) = @_;
   my ($id, $subTopic, $subName, $subId, $firstSubId, @subTopics);

   $id = $topic->attribute('id');

   @subTopics = $topic->nodeList('topic', 1);  # children only
   if (@subTopics) {
      $result .= "n$id [label=\"";
      foreach $subTopic (@subTopics) {
         $subName = $subTopic->attribute('name');
         $subId = $subTopic->attribute('id');
         $result .= "<f$subId>$subName ($subId)|";
      }
      chop($result);
      $result .= "\"];\n";

      foreach $subTopic (@subTopics) {
         generateGraph($subTopic, $id);
      }

      $firstSubId = $topic->child('topic')->attribute('id');
      if ($parentId) {
         $result .= "n$parentId:f$id -> n$id:f$firstSubId\n";
      }
   }
}


# sub generateDocBook:

# Recursively generates DocBook documentation for the topics.

sub generateDocBook ($$)
{
   my ($topic, $level) = @_;
   my ($name, $id, $summary, $subTopic, $sectionId, $child, @subTopics, 
       @children);

   $name = $topic->attribute('name');
   $id = $topic->attribute('id');
   $summary = $topic->child('summary');
   if (!$name) {
      $name = 'UNKNOWN';
   }
   if (!defined($id)) {
      $id = '999999999';
   }

   $sectionId = "$name$id";
   $sectionId =~ s/\s//g;  # strip spaces
   $sectionId =~ s|/||g;  # strip '/'

   $result .= <<EOT;

<section id=\"$sectionId\">
  <title>$name ($id)</title>

EOT

   if ($summary) {
      @children = $summary->children();
      foreach $child (@children) {
         $result .= $child->generateXml();
      }
   }

   @subTopics = $topic->nodeList('topic', 1);  # children only
   foreach $subTopic (@subTopics) {
      generateDocBook($subTopic, $level + 1);
   }

   $result .= <<EOT;

</section>
EOT

}


sub makeSedFilter ()
{
   my ($filter, $outFile);

   $filter = <<'EOT';
1,${
a\
.la
s/|/\
    /
s/|/\
        /
s/|/\
            /
s/|/\
                /
}
EOT

   $outFile = "filter";
   open(OUTFILE, ">$outFile") or die("can't open $outFile\n");
   print(OUTFILE $filter);
   close(OUTFILE) or die("can't close $outFile\n");
}


sub makeGroffMacros ()
{
   my ($macros, $outFile);

   $macros = <<'EOT';
.nr C 3      \" number of label columns 
.nr R 8      \" number of label rows
.nr W 6.6c   \" label width (centre to centre spacing)
.nr D 3.51c  \" label depth (centre to centre spacing)
.nr V 1.8c   \" effective vertical offset to top edge of labels
.nr H 0.7c   \" horizontal offset to left edge of labels
.nr A 0      \" 0 = labels left adjusted, 1 = centred

.nr c 1      \" label column counter
.nr r 1      \" label row counter
.nr w 0      \" label text width
.nr d 0      \" label text depth
.nr v 0      \" vertical offset to label top edge
.nr h 0      \" horizontal offset to label left edge
.nr f 0      \" first page boolean
.ps 12       \" select point size
.vs 12       \" select leading

.de la       \" define label macro - call follows text content

\"On entry, valid column and row values for the next label to be printed.
\"Calculate horizontal and vertical offsets. Move to new page before printing
\"if appropriate. Increment column/row values ready for next cycle. Calculate 
\"text sizes, and offsets into label space. Move to top left corner of text area
\"ready for printing.

.di                            \" end diversion - now available for measurement
.nr h \\nHu+(\\nWu*(\\nc-1))   \" set horizontal offset for this label
.nr v \\nVu+(\\nDu*(\\nr-1))   \" set vertical offset for this label
.if (\\nc=1)&(\\nr=1) \{       \" if column count = 1 and row count = 1
.ie \\nf=0 .nr f 1             \"    if not first page
.el 'bp \}                     \"       begin next page
.nr c \\nc+1                   \" increment column count for next label
.if \\nc>\\nC \{               \" if column count > number of columns
.nr c 1                        \"    reset column count for next label
.nr r \\nr+1                   \"    increment row count for next label
.if \\nr>\\nR \{               \"    if row count > number of rows
.nr r 1 \} \}                  \"       reset row count for next label
.nr v \\nvu+((\\nDu-\\n(dn)/2) \" adjust vertical position
.sp |\\nvu \}                  \" go to vertical position
.ie \\nA=0 \{                  \" if left adjustment required
.nr h \\nhu+((\\nWu-\\n(dl)/2) \"    adjust horizontal position
.in |\\nhu                     \"    go to horizontal position
.xx \}                         \"    print label
.el \{                         \" else centred
.in |\\nhu                     \"    set indent to label left edge
.ll |\\nhu+\\nWu               \"    set line length to label right edge
.ce 10                         \"    centre maximum expected lines
.xx                            \"    print label
.ce 0 \}                       \"    cancel unused centred lines
.in 0                          \" reset indent for diversion
.di xx                         \" restart diversion
..                             \" end macro definition

.fam P
.ps 12
.nf                            \" no fill mode
.po 0                          \" zero page offset
.pl 35c                        \" page length
.in 0                          \" zero indent required for diversion
.di xx                         \" divert label text into macro for measurement
EOT

   $outFile = "macros";
   open(OUTFILE, ">$outFile") or die("can't open $outFile\n");
   print(OUTFILE $macros);
   close(OUTFILE) or die("can't close $outFile\n");
}


return 1;
