package DocBook::Html::Sections;

# Processes DocBook sectioning elements which give documents their primary 
# structure: part, book, article, section, sect1, sect2, sect3, sect4, sect5, 
# para.

# SECTION DATA DEFINITIONS
#
# Level: The root section is at level 1, child sections are at level 2, and so 
# on. Level 0 is reserved for system use.
#
# Depth: The depth of a section is the depth of the section sub-tree rooted at
# the section. A section with depth 0 has no section children, with depth 1 has
# section children but no section grandchildren, and so on. The depth of a
# section can change as the tree is built.

# $Id: Sections.pm,v 1.7 2005/10/02 09:54:16 lme Exp $


use strict;

BEGIN {
   require Exporter;
   our @ISA = qw(Exporter);
   our @EXPORT = 
          qw($htmlDirectory $rootPrefix1 $rootPrefix2 $rootPrefix3 $rootPrefix4
             $rootPrefix5 $rootPrefix6 $sectionIndex $sectionLevel $isSection 
             $sectionNumbering $textNavigation $startingPage $currentPageId 
             $useLeafGroups $pageLimit $mapLimit $largestHeading 
             $smallestHeading $showNavigationMap $showHomePageMap 
	     $pathItemSeparator $menuItemSeparator @sectionIds @sectionTitle
	     @sectionIndices @sectionNumbers @sectionData @pageIds
             @globalPageIds @pageSectionLevel %isAtPageHead section); }

use Text::Wrap;
use LME::XmlNode;
use Publish qw(processChildren %dispatchTable @tagStack);
use DocBook::Html::Main qw(%pageById %xrefById);
use DocBook::Html::Footnotes qw(%footnoteData);


sub makeHeader($);
sub makeFooter($);
sub makePreviousLink($);
sub makeNextLink($);
sub makeTopLink($);
sub makeHomeLink($$);
sub makeMapLink();
sub makeMap($$);
sub makeMapPage($);


# global variables
our ($sectionLevel, $maxLevel, $htmlDirectory, $rootPrefix1, $rootPrefix2,
     $rootPrefix3, $rootPrefix4, $rootPrefix5, $rootPrefix6, $pageLimit,
     $mapLimit, $sectionIndex, $isSection, $sectionNumbering, $textNavigation, 
     $startingPage, $pageIndex, $largestHeading, $smallestHeading,
     $useLeafGroups, $showNavigationMap, $showHomePageMap, $pathItemSeparator,
     $menuItemSeparator, $pageMap, @sectionIds, @sectionTitle,
     @sectionCommands, @sectionIndices, @sectionNumbers, $currentPageId,
     @pageIds, @globalPageIds, @sectionData, @pageSectionLevel, %isAtPageHead);


sub initialize ()
{
   my ($i);

   # build dispatch table
   $dispatchTable{book} = \&book;
   $dispatchTable{article} = \&book;
   $dispatchTable{part} = \&section;
   $dispatchTable{chapter} = \&section;
   $dispatchTable{section} = \&section;
   $dispatchTable{sect1} = \&section;
   $dispatchTable{sect2} = \&section;
   $dispatchTable{sect3} = \&section;
   $dispatchTable{sect4} = \&section;
   $dispatchTable{sect5} = \&section;
   $dispatchTable{simplesect} = \&section;
   $dispatchTable{para} = \&para;

   # initialize variables
   @sectionCommands = 
      qw(part chapter section subsection subsubsection paragraph subparagraph);

   # user parameters
   if (!defined($htmlDirectory)) {
      $htmlDirectory = '../html';
   }
   if (!defined($rootPrefix1)) {
      $rootPrefix1 = '';
   }
   if (!defined($rootPrefix2)) {
      $rootPrefix2 = '';
   }
   if (!defined($rootPrefix3)) {
      $rootPrefix3 = '';
   }
   if (!defined($rootPrefix4)) {
      $rootPrefix4 = '';
   }
   if (!defined($rootPrefix5)) {
      $rootPrefix5 = '';
   }
   if (!defined($rootPrefix6)) {
      $rootPrefix6 = '';
   }
   if (!defined($pageLimit)) {
      $pageLimit = 3;
   }
   if (!defined($mapLimit)) {
      $mapLimit = 2;
   }
   if (!defined($sectionNumbering)) {
      $sectionNumbering = 1;
   }
   if (!defined($useLeafGroups)) {
      $useLeafGroups = 1;
   }
   if (!defined($textNavigation)) {
      $textNavigation = 1;
   }
   if (!defined($largestHeading)) {
      $largestHeading = 2;
   }
   if (!defined($smallestHeading)) {
      $smallestHeading = 5;
   }
   if (!defined($showNavigationMap)) {
      $showNavigationMap = 1;
   }
   if (!defined($showHomePageMap)) {
      $showHomePageMap = 0;
   }
   if (!defined($pathItemSeparator)) {
      $pathItemSeparator = " &gt; ";
   }
   if (!defined($menuItemSeparator)) {
      $menuItemSeparator = "  ";
   }
   $pathItemSeparator = "<span class=\"separator\">$pathItemSeparator</span>";
   $menuItemSeparator = "<span class=\"separator\">$menuItemSeparator</span>";

   $maxLevel = 10;
   $sectionLevel = 0;
   $sectionIndex = 0;
   $isSection = 0;
   $startingPage = 0;
   $pageIndex = 0;
   $currentPageId = 0;
   @sectionIds = ();
   @sectionIndices = ();
   @sectionNumbers = ();
   @pageIds = ();
   @globalPageIds = ();
   @pageSectionLevel = ();
   @sectionData = ();
   %isAtPageHead = ();  # identifies a section which starts a new page
   foreach $i (0 .. $maxLevel - 1) {
      $sectionNumbers[$i] = 0;
   }
   system("rm -f $htmlDirectory/*.html");
};


# sub book:
# Also handles article.

sub book ($)
{
   my ($node) = @_;
   my ($result, $sectionId, $prefix, $postfix, $fileName, $section, $pageId,
       $indent, $contents, $linkId, $line, $linkend, $map, $mapSubstituted,
       $index, $sectionTitle, $sectionNumber, $xref);
   
   $startingPage = 1;
   $isSection = 1;
   $sectionIndex++;
   $sectionLevel++;
   $sectionNumbers[$sectionLevel]++;
   if ($sectionLevel == 1) {
      $sectionId = 'index';
   } else {  # should never come here!
      $sectionId = $node->attribute('id');
      if (!$sectionId) {
         $sectionId = "s$sectionIndex";
      }
   }

   # build section data structure
   $sectionData[$sectionIndex]{id} = $sectionId;
   $sectionData[$sectionIndex]{level} = $sectionLevel;
   $sectionData[$sectionIndex]{depth} = 0;
   $sectionData[$sectionIndex]{ancestors} = [@sectionIndices];
   $sectionData[$sectionIndex]{pageId} = $sectionId;
   $sectionData[$sectionIndex]{pageIndex} = $pageIndex;
   $sectionData[$sectionIndex]{number} = $sectionNumbers[$sectionLevel];
   $pageById{$sectionId} = $sectionId;

   # push values onto stacks
   unshift(@sectionIndices, $sectionIndex);
   unshift(@sectionIds, $sectionId);
   unshift(@pageIds, $sectionId);
   unshift(@pageSectionLevel, $sectionLevel);

   # maintain current page id
   $currentPageId = $pageIds[0];

   # build global page id array
   $globalPageIds[$pageIndex] = $sectionId;

   $startingPage = 1;  # to suppress title in Title.pm
   $result = processChildren($node);

   # build html prefix
   $prefix = <<EOT;
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
  "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html>
<head>
<title>$sectionTitle[0]</title>
<link rel="stylesheet" type="text/css" href="imports/style.css"/>
</head>
<body>
EOT

   $index = $sectionIndices[0];
   $prefix .= "<div id=\"page-main\">\n";
   $prefix .= makeHeader($index);
   $prefix .= "<div id=\"page-contents\">\n";
   if ($showNavigationMap &&
       (!(($index == 1) && !$showHomePageMap))) {
      $pageMap = makeMap($sectionIndices[0], $mapLimit);
      if ($pageMap !~ m|^\s*<dl>\s*</dl>\s*$|) {
         $prefix .= "\n<div class=\"page-map\">\n";
         $prefix .= $pageMap;
         $prefix .= "</div>\n";
      }
   }

   # build html postfix
   $postfix .= "</div>\n";
   $postfix .= makeFooter($sectionIndices[0]);
   $postfix .= "</div>\n";

   $postfix .= <<EOT;
</body>
</html>
EOT

   # write out html file
   $fileName = "$htmlDirectory/$pageIds[0].html";
   open(FILE, ">$fileName") or die("can't open $fileName\n");
   print(FILE $prefix);
   print(FILE $result);
   print(FILE $postfix);
   close(FILE) or die("can't close $fileName\n");

   # finish book
   $sectionLevel--;

   # pop value from stacks
   shift(@sectionIndices);
   shift(@sectionIds);
   shift(@sectionTitle);
   shift(@pageIds);

   # maintain current page id
   $currentPageId = $pageIds[0];

   # make site map page
   makeMapPage(1);

   # make substitutions for place-holders in generated files
   foreach $pageId (@globalPageIds) {
      $contents = '';
      $mapSubstituted = 0;
      $fileName = "$htmlDirectory/$pageId.html";
      open(FILE, "$fileName") or die("can't open $fileName\n");
      while ($line = <FILE>) {
         # substitute header navigation map
         if (!$mapSubstituted) {
            if ($line =~ m/HEADER_MENU/) {
               $map = makeMap(1, 1);
               $map =~ s/$menuItemSeparator$//;  # strip final separator
               $line =~ s/HEADER_MENU/$map/;
               $mapSubstituted = 1;
            }
         }
         # substitute Next pointers
         while ($line =~ m/NEXT_LINK(\d+)/g) {
            $pageIndex = $1;
            $linkId = $globalPageIds[$pageIndex];
            if ($linkId) {
               $line =~ s/NEXT_LINK$pageIndex/<a href="$linkId.html">Next<\/a>/;
            } else {
               #$line =~ s/NEXT_LINK$pageIndex&nbsp;//;  # suppress link
               $line =~ s/NEXT_LINK$pageIndex$menuItemSeparator//;
                                                               # suppress link
            }
         }
         # substitute open links to section ids
         while ($line =~ m/LINK_START(.+?)LINK_END/g) {
            $linkend = $1;
            $pageId = $pageById{$linkend};
            if (!$pageId) {
               print("can't resolve $linkend\n");
            } else {
               $line =~ s/LINK_START${linkend}LINK_END/$pageId/;
            }
         }
         # resolve open xrefs
         while ($line =~ m/XREF_START(.+?)XREF_END/g) {
            $linkend = $1;
            $xref = $xrefById{$linkend};
            if (!$xref) {
               print("can't resolve $linkend xref\n");
            } else {
               $line =~ s/XREF_START${linkend}XREF_END/$xref/;
            }
         }
         $contents .= $line;
      }
      unlink($fileName) or die("can't unlink $fileName\n");
      open(FILE, ">$fileName") or die("can't open $fileName\n");
      print(FILE $contents);
      close(FILE) or die("can't close $fileName\n");
   }

   return '';
}


# sub section():
# Also handles part, chapter, sect1-5, simplesect.

sub section ($)
{
   my ($node) = @_;
   my ($result, $i, $depth, $newTopic, $prefix, $postfix, $sectionClass,
       $sectionId, $fileName, $index, $sectionTitle, $sectionNumber);
   
   $isSection = 1;
   $sectionIndex++;
   $sectionLevel++;
   $sectionNumbers[$sectionLevel]++;
   $sectionClass = $node->attribute('class');
   if (!$sectionClass) {
      $sectionClass = 'none';
   }
   $sectionId = $node->attribute('id');
   if (!$sectionId) {
      $sectionId = "s$sectionIndex";
   }

   # build section data structure
   $sectionData[$sectionIndex]{id} = $sectionId;
   $sectionData[$sectionIndex]{level} = $sectionLevel;
   $sectionData[$sectionIndex]{ancestors} = [@sectionIndices];
   $sectionData[$sectionIndex]{depth} = 0;

   # record current depths
   $depth = 0;
   foreach $i (@sectionIndices) {
      $depth++;
      if ($sectionData[$i]{depth} < $depth) {
         $sectionData[$i]{depth} = $depth;
      }
   }

   # push values onto stacks
   unshift(@sectionIndices, $sectionIndex);
   unshift(@sectionIds, $sectionId);

   # is new page required?
   if (($node->tag() eq 'part') || ($node->tag() eq 'chapter') ||
       #($sectionClass eq 'topic') || ($sectionLevel > $pageLimit)) {
       ($sectionClass eq 'topic')) {
      $startingPage = 1;  # to suppress title in Title.pm
      unshift(@pageIds, $sectionId);  # push onto stack
      unshift(@pageSectionLevel, $sectionLevel);
      $currentPageId = $pageIds[0];
      $pageIndex++;
      $globalPageIds[$pageIndex] = $sectionId;  # build global array
      $isAtPageHead{$sectionId} = 1;
   }
   $sectionData[$sectionIndex]{pageId} = $pageIds[0];
   $sectionData[$sectionIndex]{pageIndex} = $pageIndex;
   $pageById{$sectionId} = $pageIds[0];

   $result = "\n<div class=\"section\">\n";  # to handle IE Peekaboo bug
   $result .= processChildren($node);
   $result .= "</div>\n";

   $sectionNumbers[$sectionLevel + 1] = 0;

   if (($node->tag() eq 'part') || ($node->tag() eq 'chapter') ||
       #($sectionClass eq 'topic') || ($sectionLevel > $pageLimit)) {
       ($sectionClass eq 'topic')) {
      # build html prefix
      $prefix = <<EOT;
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
  "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html>
<head>
<title>$sectionTitle[0]</title>
<link rel="stylesheet" type="text/css" href="imports/style.css"/>
</head>
<body>
EOT

      $index = $sectionIndices[0];
      $prefix .= "<div id=\"page-main\">\n";
      $prefix .= makeHeader($index);
      $prefix .= "<div id=\"page-contents\">\n";
      $sectionTitle = $sectionData[$index]{title};
      $sectionNumber = $sectionData[$index]{number};
      if ($sectionNumbering) {
         $prefix .= "<h$largestHeading>$sectionNumber&nbsp;";
         $prefix .= "$sectionTitle</h$largestHeading>\n";
      } else {
         $prefix .= "<h$largestHeading>$sectionTitle</h$largestHeading>\n";
      }
      if ($showNavigationMap &&
          (!(($index == 1) && !$showHomePageMap))) {
         $pageMap = makeMap($sectionIndices[0], $mapLimit);
         if ($pageMap !~ m|^\s*<dl>\s*</dl>\s*$|) {
            $prefix .= "\n<div class=\"page-map\">\n";
            $prefix .= $pageMap;
            $prefix .= "</div>\n";
         }
      }

      # build html postfix
      $postfix = makeFootnotes();
      $postfix .= "</div>\n";
      $postfix .= makeFooter($sectionIndices[0]);
      $postfix .= "</div>\n";

      $postfix .= <<EOT;
</body>
</html>
EOT

      # write out html file
      $fileName = "$htmlDirectory/$pageIds[0].html";
      open(FILE, ">$fileName") or die("can't open $fileName\n");
      print(FILE $prefix);
      print(FILE $result);
      print(FILE $postfix);
      close(FILE) or die("can't close $fileName\n");

      # pop value from stack
      shift(@pageIds);
      shift(@pageSectionLevel);

      # maintain current page id
      $currentPageId = $pageIds[0];

      $result = '';
   }

   # finish section
   $sectionLevel--;

   # pop value from stacks
   shift(@sectionIndices);
   shift(@sectionIds);
   shift(@sectionTitle);

   return $result;
}


sub para ($)
{
   my ($node) = @_;
   my ($id, $result);
   
   # now handled at top level
   #$id = $node->attribute('id');
   #if ($id) {
   #   $pageById{$id} = $currentPageId;
   #}

   $result = "<p>";

   $result .= processChildren($node);

   $result .= "</p>\n";

   return $result;
}


# UTILITIES

sub makeHeader ($)
{
   my ($index) = @_;
   my ($sectionId, $sectionNumber, $sectionTitle, $header, $rootIndex, 
       $rootTitle);

   if ($index > 1) {
      $rootIndex = $sectionData[$index]{ancestors}[-1];
   } else {
      $rootIndex = $index;
   }
   $rootTitle = $sectionData[$rootIndex]{title};
   $sectionId = $sectionData[$index]{id};
   $sectionTitle = $sectionData[$index]{title};
   $sectionNumber = $sectionData[$index]{number};
   $header = "<div id=\"page-header\">\n";
   $header .= "<div id=\"header-path\">\n";
   if ($index == 1) {
      $header .= "Home\n";
   } else {
      $header .= makeHomeLink($index, 0);
      $header .= $pathItemSeparator;
      $header .= makePath($sectionIndices[0]);
   }
   $header .= "\n</div>\n";
   $header .= "<div id=\"header-nav\">\n";
   if ($index > 1) {
      $header .= makePreviousLink($index);
      $header .= $menuItemSeparator;
   }
   $header .= makeNextLink($index);
   $header .= $menuItemSeparator;
   $header .= makeMapLink();
   $header .= "\n</div>\n";
   $header .= "<div id=\"root-prefix\">\n";
   $header .= makePrefixLinks();
   $header .= "</div>\n";
   $header .= "<div id=\"root-title\">\n";
   $header .= makeHomeLink($index, 1);
   $header .= "</div>\n";
   $header .= "<div id=\"header-menu\">\n";
   $header .= "HEADER_MENU";
   $header .= "</div>\n";
   $header .= "</div>\n";

   return $header;
}


sub makeFooter ($)
{
   my ($index) = @_;
   my ($footer, $path, $nav);

   if ($index == 1) {
      $path = "Home\n";
   } else {
      $path = makeHomeLink($index, 0);
      $path .= $pathItemSeparator;
      $path .= makePath($sectionIndices[0]);
   }

   $nav = makeTopLink($index);
   $nav .= $menuItemSeparator;
   if ($index > 1) {
      $nav .= makePreviousLink($index);
      $nav .= $menuItemSeparator;
   }
   $nav .= makeNextLink($index);
   $nav .= $menuItemSeparator;
   $nav .= makeMapLink();

   $footer = <<EOT;
<div id="page-footer">
<div id="footer-path">
$path
</div>
<div id="footer-nav">
$nav
</div>
</div>
EOT

   return $footer;
}


sub makePreviousLink ($)
{
   my ($index) = @_;
   my ($sectionId, $link, $pageId, $pageIndex, $fileName);

   $pageIndex = $sectionData[$index]{pageIndex};
   $fileName = "$globalPageIds[$pageIndex - 1].html";
   $link = "<a href=\"$fileName\">Previous</a>";

   return $link;
}
 

sub makeNextLink ($)
{
   my ($index) = @_;
   my ($link, $pageIndex);

   $pageIndex = $sectionData[$index]{pageIndex} + 1;
   $link = "NEXT_LINK$pageIndex";

   return $link;
}
 

sub makeTopLink ($)
{
   my ($index) = @_;
   my ($link, $pageId, $fileName);

   $pageId = $sectionData[$index]{pageId};
   $fileName = "$pageId.html";
   $link = "<a href=\"$fileName\">Top</a>";

   return $link;
}
 

sub makeHomeLink ($$)
{
   my ($index, $displayTitle) = @_;
   my ($link, $fileName, $rootIndex, $rootTitle);

   $fileName = "$globalPageIds[0].html";
   if ($index > 1) {
      $rootIndex = $sectionData[$index]{ancestors}[-1];
   } else {
      $rootIndex = $index;
   }
   $rootTitle = $sectionData[$rootIndex]{title};
   if ($displayTitle) {
      $link = "<a href=\"$fileName\">$rootTitle</a>";
   } else {
      $link = "<a href=\"$fileName\">Home</a>";
   }

   return $link;
}
 

sub makeMapLink ()
{
   my ($sectionId, $link, $pageId, $fileName);

   $fileName = "sitemap.html";
   $link = "<a href=\"$fileName\">Index</a>";

   return $link;
}


sub makePrefixLinks ()
{
   my ($name, $url, $links);

   $links = '';
   if ($rootPrefix1) {
      ($name, $url) = ($rootPrefix1 =~ m/([^,]*), *(.*)/);
      $links .= "<a href=\"$url\">$name</a>";
   }
   if ($rootPrefix2) {
      ($name, $url) = ($rootPrefix2 =~ m/([^,]*), *(.*)/);
      $links .= "$pathItemSeparator<a href=\"$url\">$name</a>";
   }
   if ($rootPrefix3) {
      ($name, $url) = ($rootPrefix3 =~ m/([^,]*), *(.*)/);
      $links .= "$pathItemSeparator<a href=\"$url\">$name</a>";
   }
   if ($rootPrefix4) {
      ($name, $url) = ($rootPrefix4 =~ m/([^,]*), *(.*)/);
      $links .= "$pathItemSeparator<a href=\"$url\">$name</a>";
   }
   if ($rootPrefix5) {
      ($name, $url) = ($rootPrefix5 =~ m/([^,]*), *(.*)/);
      $links .= "$pathItemSeparator<a href=\"$url\">$name</a>";
   }
   if ($rootPrefix6) {
      ($name, $url) = ($rootPrefix6 =~ m/([^,]*), *(.*)/);
      $links .= "$pathItemSeparator<a href=\"$url\">$name</a>";
   }

   return $links;
}
 

sub makeMap ($$)
{
   my ($startIndex, $mapLimit) = @_;
   my ($index, $lastIndex, $sectionId, $map, $fileName, $section, 
       $level, $depth, $previousLevel, $previousDepth, $startLevel, $pageId, 
       $sectionTitle, $sectionNumber, $leafLevel, $leafGroup, $useList);

   # initialize
   $useList = !(($startIndex == 1) && ($mapLimit == 1));
   if ($useList) {
      $map = "<dl>\n";
   }
   $previousDepth = 0;
   $leafGroup = 0;
   $index = $startIndex;
   $startLevel = $sectionData[$index]{level};
   $previousLevel = $startLevel;
   $lastIndex = scalar(@sectionData) - 1;

   # process @sectionData array
   while ($index <= $lastIndex) {
      $level = $sectionData[$index]{level};

      # stop processing?
      if (($index > $startIndex) && ($level < $startLevel)) {
         return $map;
      }

      # skip sections below map limit
      if ($level - $startLevel > $mapLimit) {
         $index++;
         next;
      }

      # control indenting
      if ($index > $startIndex + 1) {
         if ($level > $previousLevel) {
            foreach ($previousLevel .. $level - 1) {
               if ($useList) {
                  $map .= "<dd><dl>\n";
               }
            }
         } elsif ($level < $previousLevel) {
            foreach ($level .. $previousLevel - 1) {
               if ($leafGroup) {  # terminate leaf group
                  if ($useList) {
                     # strip final menu item separator
                     $map =~ s/$menuItemSeparator$//;
                     $map .= "\n</dt></dl></dd>\n";
                  }
                  $leafGroup = 0;
               } else {
                  if ($useList) {
                     $map .= "</dl></dd>\n";
                  }
               }
            }
         }
      }
   
      # get section data values
      $sectionId = $sectionData[$index]{id};
      $pageId = $sectionData[$index]{pageId};
      $sectionNumber = $sectionData[$index]{number};
      $sectionTitle = $sectionData[$index]{title};
      $fileName = "$pageId.html";
      $depth = $sectionData[$index]{depth};

      # build navigation map, suppressing start title
      if ($index > $startIndex) {

         # start a leaf group?
         if ($useLeafGroups && !$leafGroup) {
            if ((($depth == 0) && ($previousDepth == 1)) ||
                ($level == $startLevel + $mapLimit)) {
               $leafGroup = 1;
               $leafLevel = $level;
               if ($useList) {
                  $map .= "<dt>\n";
               }
            }
         }

         # process leaf group section
         if (($leafGroup) && ($level == $leafLevel)) {
            $sectionTitle =~ s/ /&nbsp;/g;
            if ($sectionNumbering) {
               $map .= "$sectionNumber&nbsp;";
            }
            if (($startIndex == 1) && ($mapLimit == 1)) {
               $map .= "<a href=\"$fileName\">";
            } else {
               $map .= "<a href=\"$fileName#$sectionId\">";
            }
            $map .= "$sectionTitle</a>$menuItemSeparator";
         }

         # process normal section
         if (!$leafGroup) {
            # strip final separator if previous section was leaf group
            $map =~ s/$menuItemSeparator\n$//;

            if ($useList) {
               $map .= "<dt>\n";
            }
            if ($sectionNumbering) {
               $map .= "$sectionNumber&nbsp;";
            }
            if (($startIndex == 1) && ($mapLimit == 1)) {
               $map .= "<a href=\"$fileName\">";
            } else {
               $map .= "<a href=\"$fileName#$sectionId\">";
            }
            # NEW WORK
            if ($isAtPageHead{$sectionId}) {
               $map .= "<b>$sectionTitle</b></a>\n";
            } else {
               $map .= "$sectionTitle</a>\n";  # ORIG
            }
            if ($useList) {
               $map .= "</dt>\n";
            }
         }
      }

      $index++;
      $previousLevel = $level;
      $previousDepth = $depth;
   }
   if ($leafGroup) {  # terminate final leaf group
      if ($useList) {
         # strip final menu item separator
         $map =~ s/$menuItemSeparator$//;
         $map .= "\n</dt></dl></dd>\n";
      }
      $leafGroup = 0;
   } 
   if ($useList) {
      $map .= "</dl>\n";
   }

   return $map;
}


sub makeMapPage ($)
{
   my ($index) = @_;
   my ($result, $fileName, $homeLink, $rootIndex, $rootTitle, $map, $siteMap);

   $rootIndex = 1;
   $rootTitle = $sectionData[$rootIndex]{title};
   $homeLink = makeHomeLink($index, 0);
   $map = makeMap(1, 1);
   $map =~ s/$menuItemSeparator$//;  # strip final separator
   $siteMap = makeMap(1, 10);

   $result = <<EOT;
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
  "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html>
<head>
<title>Index</title>
<link rel="stylesheet" type="text/css" href="imports/style.css"/>
</head>
<body>
<div id="page-main">
<div id="page-header">
<div id="header-path">Index</div>
EOT

   $result .= "<div id=\"root-prefix\">\n";
   $result .= makePrefixLinks();
   $result .= "</div>\n";
   $result .= "<div id=\"root-title\">\n";
   $result .= makeHomeLink($index, 1);
   $result .= "</div>\n";


   $result .= <<EOT;
<div id="header-menu">
$map
</div> 
</div>
<div id="page-contents">
<h$largestHeading>Index</h$largestHeading>
<b>$homeLink</b>
<br/>
<div class="site-map">
$siteMap
</div>
</div>
<div id="page-footer">
<div id="footer-path">
Index
</div>
<div id="footer-nav">
&nbsp;
</div>
</div>
</div>
</body>
</html>
EOT

   # write out html file
   $fileName = "$htmlDirectory/sitemap.html";
   open(FILE, ">$fileName") or die("can't open $fileName\n");
   print(FILE $result);
   close(FILE) or die("can't close $fileName\n");
}


sub makePath ()
{
   my ($index) = @_;
   my ($path, $pathIndex, $pageId, $fileName, @pathIndices);

   $path = '';
   @pathIndices = @{$sectionData[$index]{ancestors}};
   foreach $pathIndex (reverse(@pathIndices)) {
      if ($pathIndex > 1) {
         $pageId = $sectionData[$pathIndex]{pageId};
         #if ($pathIndex == 1) {
         #   $fileName = "index.html";
         #} else {
            $fileName = "$pageId.html";
         #}
         if ($sectionNumbering) {
            $path .= "$sectionData[$pathIndex]{number}&nbsp;";
         }
         $path .= "<a href=\"$fileName#$sectionData[$pathIndex]{id}\">";
         $path .= "$sectionData[$pathIndex]{title}</a>\n";
         $path .= "$pathItemSeparator";
      }
   }

   # add current section name to path
   if ($sectionNumbering) {
      $path .= "$sectionData[$index]{number}&nbsp;";
   }
   $path .= "$sectionData[$index]{title}\n";

   return $path;
}


sub makeFootnotes ()
{
   my ($footnotes, $footnote, $index);

   if ($footnoteData{$currentPageId}) {
      $footnotes = "<h$smallestHeading>Footnotes</h$smallestHeading>\n";
      $footnotes .= "<table>\n";
      $index = 1;
      foreach $footnote (@{$footnoteData{$currentPageId}}) {
         $footnotes .= "<tr><td>[$index]</td><td>$footnote</td></tr>\n";
         $index++;
      }
      $footnotes .= "</table>\n";
   }

   return $footnotes;
}


return 1;
