package DocBook::Latex::Structure;

# Processes DocBook elements which give documents their primary structure:
# part, book, article, section, sect1, sect2, sect3, sect4, sect5, para.

# Latex treats the higher level sectioning commands differently to paragraph
# and subparagraph (defined as 'run-in types'). See Kopka & Daly p348.  For
# normal DocBook processing we need to change this so all sectioning commands
# behave the same. Hence renewcommand...startsection. This also fixes a bug in
# Latex, where a longtable as the only element in a paragraph element comes 
# before the paragraph heading! This is nothing to do with floats, because 
# longtables don't float.

# $Id: Structure.pm,v 1.2 2005/07/10 11:31:47 lme Exp $


use strict;

BEGIN {
   require Exporter;
   our @ISA = qw(Exporter);
   # our @EXPORT_OK = qw();
}

use LME::XmlNode;
use Publish qw(processChildren %dispatchTable @tagStack);
use DocBook::Latex::Main qw($preamble $postfix);

# provate global variables
our ($sectionDepth, @sectionCommands);


sub initialize ()
{
   # build dispatch table
   $dispatchTable{part} = \&part;
   $dispatchTable{book} = \&book;
   $dispatchTable{article} = \&article;
   $dispatchTable{chapter} = \&section;
   $dispatchTable{section} = \&section;
   $dispatchTable{sect1} = \&section;
   $dispatchTable{sect2} = \&section;
   $dispatchTable{sect3} = \&section;
   $dispatchTable{sect4} = \&section;
   $dispatchTable{sect5} = \&section;
   $dispatchTable{simplesect} = \&section;
   $dispatchTable{para} = \&para;

   # define Latex section command strings and initial depth
   @sectionCommands = 
      qw(chapter section subsection subsubsection paragraph subparagraph
         subparagraph subparagraph subparagraph subparagraph subparagraph);
   $sectionDepth = 0;

   # build latex preamble
   $preamble .= <<EOT;

% structure
\\setcounter{secnumdepth}{5}
%\\setcounter{secnumdepth}{0}  %inhibit - 
                              % works on the page, but pdf bookmarks don't
\\setcounter{tocdepth}{5}

% the following definitions are based on book.cls
\\makeatletter
\\def\\\@makechapterhead#1{%
  {\\parindent \\z\@ \\raggedright \\normalfont
    \\ifnum \\c\@secnumdepth >\\m\@ne
      \\if\@mainmatter
        \\par\\nobreak
      \\fi
    \\fi
    \\interlinepenalty\\\@M
    \\LARGE \\bfseries \\thechapter. #1\\par\\nobreak
    %\\vspace{2.3ex \\\@plus.2ex}
    \\vspace{1.0ex \\\@plus.2ex}
  }}
  \\renewcommand{\\section}{\\\@startsection {section}{1}{\\z\@}%
                   %{-3.5ex \\\@plus -1ex \\\@minus -.2ex}%
                   {-2.5ex \\\@plus -1ex \\\@minus -.2ex}%
                   %{2.3ex \\\@plus.2ex}%
                   {1.0ex \\\@plus .2ex}%
                   {\\normalfont\\Large\\bfseries}}
  \\renewcommand{\\subsection}{\\\@startsection{subsection}{2}{\\z\@}%
                   %{-3.25ex\\\@plus-1ex \\\@minus -.2ex}%
                   {-2.25ex\\\@plus-1ex \\\@minus -.2ex}%
                   %{1.5ex \\\@plus.2ex}%
                   {1.0ex \\\@plus.2ex}%
                   {\\normalfont\\large\\bfseries}}
  \\renewcommand{\\subsubsection}{\\\@startsection{subsubsection}{3}{\\z\@}%
                   {-3.25ex\\\@plus-1ex \\\@minus -.2ex}%
                   %{1.5ex \\\@plus.2ex}%
                   {1.0ex \\\@plus.2ex}%
                   {\\normalfont\\normalsize\\bfseries}}
  \\renewcommand{\\paragraph}{\\\@startsection{paragraph}{4}{\\z\@}\%
                   {-3.25ex \\\@plus-1ex \\\@minus-.2ex}\%
                   {0.5ex \\\@plus0.2ex}\%
                   {\\normalfont\\normalsize\\bfseries}}
  \\renewcommand{\\subparagraph}{\\\@startsection{subparagraph}{5}{\\z\@}\%
                   {-3.25ex \\\@plus-1ex \\\@minus-.2ex}\%
                   {0.5ex \\\@plus0.2ex}\%
                   {\\normalfont\\normalsize\\bfseries}}

% from article.cls, used for articles only
\\def\\\@maketitle{%
  \\newpage
  \\null
  %\\vskip 2em%
  \\begin{center}%
  \\let \\footnote \\thanks
    {\\LARGE \\bfseries \\\@title \\par}%
    \\vskip 1.5em%
    %{\\large
    %  \\lineskip .5em%
    %  \\begin{tabular}[t]{c}%
    %    \\\@author
    %  \\end{tabular}\\par}%
    %\\vskip 1em%
    {\\large \\\@date}%
  \\end{center}%
  \\par
  \\vskip 1.5em}

\\makeatother
EOT

};


sub book ($)
{
   my ($node) = @_;
   my ($result);
   
   $result = $preamble;

   $result .= processChildren($node);

   $result .= $postfix;

   return $result;
}


sub article ($)
{
   my ($node) = @_;
   my ($result);
   
   $sectionDepth++;

   # fix preamble
   $preamble =~ s/{book}/{article}/;

   $result = $preamble;

   $result .= processChildren($node);

   $result .= $postfix;

   $sectionDepth--;

   return $result;
}


sub section ($)
{
   my ($node) = @_;
   my ($result, $id);
   
   $id = $node->attribute('id');

   $result = "\\$sectionCommands[$sectionDepth]";
   $sectionDepth++;

   $result .= processChildren($node);

   $result .= "\n\\label{$id}\n";

   $sectionDepth--;

   return $result;
}


sub para ($)
{
   my ($node) = @_;
   my ($result);
   
   $result = "\n";

   $result .= processChildren($node);

   $result .= "\n";

   return $result;
}


return 1;
