package DocBook::Rtf::Tables;

# Processes DocBook table elements: table, informaltable, tgroup, thead, tfoot, 
# tbody, colspec, spanspec, row, entry. Both the CALS and Exchange table 
# models are handled, with some restrictions.

# $Id: Tables.pm,v 1.1 2005/07/10 11:26:33 lme Exp $


use strict;

BEGIN {
   require Exporter;
   our @ISA = qw(Exporter);
   our @EXPORT_OK = qw($tableColumnSpacing %table);
}

use LME::XmlNode;
use Publish qw(processChildren %dispatchTable @tagStack);
use DocBook::Rtf::Main qw($preamble);


# public global variables
our ($tableColumnSpacing, %table);

# private global variables
our ($tgroupIndex, $colspecIndex, $spanspecIndex, $rowIndex, $columnIndex, 
     $entryIndex, $maxColspecIndex, $savedResult);


sub initialize ()
{
   # build dispatch table
   $dispatchTable{table} = \&table;
   $dispatchTable{informaltable} = \&table;
   $dispatchTable{tgroup} = \&tgroup;
   $dispatchTable{thead} = \&thead;
   $dispatchTable{tfoot} = \&tfoot;
   $dispatchTable{tbody} = \&tbody;
   $dispatchTable{colspec} = \&colspec;
   $dispatchTable{spanspec} = \&spanspec;
   $dispatchTable{row} = \&row;
   $dispatchTable{entry} = \&entry;

   # initialize element variables
   %table = ();
   $tgroupIndex = 0;
   $colspecIndex = 0;
   $spanspecIndex = 0;
   $rowIndex = 0;
   $columnIndex = 0;
   $entryIndex = 0;
   $maxColspecIndex = 0;
   $savedResult = '';

   # table constants
   $tableColumnSpacing = 2.0;  # mm

   # need a method to set parameters from a style sheet, before generating the 
   # prefix

   # add to rtf preamble if required
   $preamble .= <<EOT;
EOT

}


# sub table:
# Processes table and informaltable. Default values can also come from a
# stylesheet though this is not currently implemented. Note that we divert 
# $result so we can deal more intelligently with table entries.

sub table ($)
{
   my ($node) = @_;
   my ($result, $colsep, $frame, $pgwide, $rowsep, $tgroup, $thead, $tbody, 
       $row, $entry, $colspec, $colwidth, $firstIndex, $lastIndex, 
       $colspecCount, $entries, $clines, $morerows, $tag);

   $tag = $tagStack[0];
   $colsep = $node->attribute('colsep');
   $frame = $node->attribute('frame');
   $pgwide = $node->attribute('pgwide');
   $rowsep = $node->attribute('rowsep');

   %table = ();
   if ($tag eq 'informaltable') {
      $table{name} = '';
   }
   $table{colsep} = $colsep ? $colsep : 1;
   $table{frame} = $frame ? $frame : 'all';
   $table{pgwide} = $pgwide ? $pgwide : 0;
   $table{rowsep} = $rowsep ? $rowsep : 1;
   $tgroupIndex = 0;
   $colspecIndex = 0;
   $spanspecIndex = 0;
   $rowIndex = 0;
   $colspecIndex = 0;
   $entryIndex = 0;

   $result = processChildren($node);

   # output table
   $result .= "\\begin{longtable}{|";
   foreach $tgroup (@{$table{tgroups}}) {
      foreach $colspec (@{$$tgroup{colspecs}}) {
         $colwidth = $$colspec{colwidth};
         $result .= "p{$colwidth mm}|";
      }
      $result .= "}\n";
      if ($table{name}) {
         #$result .= "\\caption*{$table{name}}\n";  # without numbering
         $result .= "\\caption{$table{name}}\n";  # with numbering
         $result .= "\\tabularnewline\n";
      }
      $result .= "\\hline\n";

      # thead
      $thead = $$tgroup{thead};
      if ($thead) {
         foreach $row (@{$$thead{rows}}) {
            $colspecIndex = 0;
            $entries = '';
            $clines = '';
            foreach $entry (@{$$row{entries}}) {
               $colsep = $$entry{colsep};
               $rowsep = $$entry{rowsep};
               $morerows = $$entry{morerows};
               $firstIndex = $$entry{firstIndex};
               $lastIndex = $$entry{lastIndex};
               $colspecCount = $lastIndex - $firstIndex + 1;
               $colwidth = 0;
               foreach $colspecIndex ($firstIndex .. $lastIndex) {
                  $colwidth += $$tgroup{colspecs}[$colspecIndex]{colwidth};
               }
               while ($colspecIndex < $firstIndex) {
                  $entries .= " & ";
                  $colspecIndex++;
                  $clines .= "\\cline{$colspecIndex - $colspecIndex} ";
               }
               if ($colspecCount == 1) {
                  $entries .= "\\bfseries \\raggedright $$entry{content} & ";
                  $colspecIndex++;
               } else {
                  # poor workaround for latex 'verbatim in multicolumn' bug
                  # all we can do is not use verbatim!
                  $$entry{content} =~ s/\\begin{verbatim}//;
                  $$entry{content} =~ s/\\end{verbatim}\n//;

                  $entries .= 
                     "\\multicolumn{$colspecCount}{|p{$colwidth mm}|}";
                  $entries .= 
                     "{\\bfseries \\raggedright $$entry{content}} & ";
                  $colspecIndex = $lastIndex + 1;
               }
               $firstIndex++;  # 1-based required
               $lastIndex++;
               if (!$morerows) {
                  $clines .= "\\cline{$firstIndex - $lastIndex} ";
               }
            }
            chop($entries);
            chop($entries);
            $entries .= "\\tabularnewline\n";
            $result .= $entries;
            $result .= $clines;
         }
         $result .= "\\endhead";
      }

      # tbody
      $tbody = $$tgroup{tbody};
      if ($tbody) {
         foreach $row (@{$$tbody{rows}}) {
            $colspecIndex = 0;
            $entries = '';
            $clines = '';
            foreach $entry (@{$$row{entries}}) {
               $colsep = $$entry{colsep};
               $rowsep = $$entry{rowsep};
               $morerows = $$entry{morerows};
               $firstIndex = $$entry{firstIndex};
               $lastIndex = $$entry{lastIndex};
               $colspecCount = $lastIndex - $firstIndex + 1;
               $colwidth = 0;
               foreach $colspecIndex ($firstIndex .. $lastIndex) {
                  $colwidth += $$tgroup{colspecs}[$colspecIndex]{colwidth};
               }
               while ($colspecIndex < $firstIndex) {
                  $entries .= " & ";
                  $colspecIndex++;
                  $clines .= "\\cline{$colspecIndex - $colspecIndex} ";
               }
               if ($colspecCount == 1) {
                  $entries .= "\\raggedright $$entry{content} & ";
                  $colspecIndex++;
               } else {
                  $entries .= 
                     "\\multicolumn{$colspecCount}{|p{$colwidth mm}|}";
                  $entries .= "{\\raggedright $$entry{content}} & ";
                  $colspecIndex = $lastIndex + 1;
               }
               $firstIndex++;  # 1-based required
               $lastIndex++;
               if (!$morerows) {
                  $clines .= "\\cline{$firstIndex - $lastIndex} ";
               }
            }
            chop($entries);
            chop($entries);
            $entries .= "\\tabularnewline\n";
            $result .= $entries;
            $result .= $clines;
         }
      }
   }
   $result .= "\\hline\n";
   $result .= "\\end{longtable}\n";

   return $result;
}


sub tgroup ($)
{
   my ($node) = @_;
   my ($result, $i, $align, $cols, $colsep, $rowsep, $tgroup, $colspec);

   $align = $node->attribute('align');
   $cols = $node->attribute('cols');
   $colsep = $node->attribute('colsep');
   $rowsep = $node->attribute('rowsep');

   $table{tgroups}[$tgroupIndex] = {};
   $tgroup = $table{tgroups}[$tgroupIndex];
   $$tgroup{align} = $align ? $align : 'left';
   $$tgroup{cols} = $cols ? $cols : 1;
   $$tgroup{colsep} = $colsep ? $colsep : $table{colsep};
   $$tgroup{rowsep} = $rowsep ? $rowsep : $table{rowsep};

   $maxColspecIndex = $cols - 1;
   foreach $colspecIndex (0..$maxColspecIndex) {
      $$tgroup{colspecs}[$colspecIndex] = {};
      $colspec = $$tgroup{colspecs}[$colspecIndex];
      $$colspec{align} = $align;
      $$colspec{char} = '';
      $$colspec{charoff} = '50';
      $$colspec{colname} = $colspecIndex + 1;
      $$colspec{colnum} = $colspecIndex + 1;
      $$colspec{colsep} = $colsep;
      $$colspec{colwidth} = '1*';
      $$colspec{rowsep} = $rowsep;
   }
   $colspecIndex = 0;

   $result = processChildren($node);

   $tgroupIndex++;

   return $result;  # null expected
}


# sub colspec:
# A set of default colspecs is created when the enclosing tgroup is processed.
# This routine does consistency checks and overwrites defaults with attributes
# values where they are defined.

sub colspec ($)
{
   my ($node) = @_;
   my ($result, $align, $char, $charoff, $colname, $colnum, $colsep, $colwidth,
       $rowsep, $value, $units, $tgroup, $colspec);

   $align = $node->attribute('align');
   $char = $node->attribute('char');
   $charoff = $node->attribute('charoff');
   $colname = $node->attribute('colname');
   $colnum = $node->attribute('colnum');
   $colsep = $node->attribute('colsep');
   $colwidth = $node->attribute('colwidth');
   $rowsep = $node->attribute('rowsep');

   # compute colwidth in mm
   # ASSUME ALWAYS DEFINED HERE, AND LIMITED CONVERSIONS FOR NOW
   ($value, $units) = ($colwidth =~ m/([\d\.]+)(\w+)/);
   if ($units eq 'cm') {
      $value *= 10;
   }
   $value -= 2*$tableColumnSpacing;
   $colwidth = $value;

   $tgroup = $table{tgroups}[$tgroupIndex];
   $colspecIndex = $colnum ? $colnum - 1 : $colspecIndex;
   $colspec = $$tgroup{colspecs}[$colspecIndex];
   $$colspec{align} = $align ? $align : $$colspec{align};
   $$colspec{char} = $char ? $char : $$colspec{char};
   $$colspec{charoff} = $charoff ? $charoff : $$colspec{charoff};
   $$colspec{colname} = $colname ? $colname : $$colspec{colname};
   $$colspec{colsep} = $colsep ? $colsep : $$colspec{colsep};
   $$colspec{colwidth} = $colwidth ? $colwidth : $$colspec{colwidth};
   $$colspec{rowsep} = $rowsep ? $rowsep : $$colspec{rowsep};

   $result = processChildren($node);

   $colspecIndex++;

   return $result;  # null expected
}


# sub spanspec:
# This is a CALS table element only, not present in the Exchange Model.
# Provided for backward compatibility with the CALS Model.

sub spanspec ($)
{
   my ($node) = @_;
   my ($result, $align, $char, $charoff, $colsep, $namest, $nameend, $rowsep, 
       $spanname, $tgroup, $spanspec);
   
   $align = $node->attribute('align');
   $char = $node->attribute('char');
   $charoff = $node->attribute('charoff');
   $colsep = $node->attribute('colsep');
   $namest = $node->attribute('namest');
   $nameend = $node->attribute('nameend');
   $rowsep = $node->attribute('rowsep');
   $spanname = $node->attribute('spanname');

   $tgroup = $table{tgroups}[$tgroupIndex];
   $$tgroup{spanspecs}[$spanspecIndex] = {};
   $spanspec = $$tgroup{spanspecs}[$spanspecIndex];
   $$spanspec{align} = $align ? $align : '';
   $$spanspec{char} = $char ? $char : '';
   $$spanspec{charoff} = $charoff ? $charoff : '';
   $$spanspec{colsep} = $colsep ? $colsep : '';
   $$spanspec{namest} = $namest ? $namest : '';
   $$spanspec{nameend} = $nameend ? $nameend : '';
   $$spanspec{rowsep} = $rowsep ? $rowsep : '';
   $$spanspec{spanname} = $spanname ? $spanname : '';

   $result = processChildren($node);

   $spanspecIndex++;

   return $result;  # null expected
}


sub thead ($)
{
   my ($node) = @_;
   my ($result, $valign, $thead);

   $valign = $node->attribute('valign');
   $table{tgroups}[$tgroupIndex]{thead} = {};
   $thead = $table{tgroups}[$tgroupIndex]{thead};
   $$thead{valign} = $valign ? $valign : 'bottom';
   $rowIndex = 0;

   $result = processChildren($node);

   return $result;  # null expected
}


# sub tfoot:
# Not allowed in the Exchange table model, so not currently processed.

sub tfoot ($)
{
   my ($node) = @_;
   my ($result);

   $result = processChildren($node);

   return '';
}


sub tbody ($)
{
   my ($node) = @_;
   my ($result, $i, $valign, $tbody);

   $valign = $node->attribute('valign');
   $table{tgroups}[$tgroupIndex]{tbody} = {};
   $tbody = $table{tgroups}[$tgroupIndex]{tbody};
   $$tbody{valign} = $valign ? $valign : 'top';
   $rowIndex = 0;

   $result = processChildren($node);

   return $result;  # null expected
}


sub row ($)
{
   my ($node) = @_;
   my ($result, $parent, $rowsep, $valign, $tgroup, $row);

   $rowsep = $node->attribute('rowsep');
   $valign = $node->attribute('valign');
   $parent = $tagStack[1];  # 'thead' | 'tbody'

   $tgroup = $table{tgroups}[$tgroupIndex];
   $$tgroup{$parent}{rows}[$rowIndex] = {};
   $row = $$tgroup{$parent}{rows}[$rowIndex];
   $$row{rowsep} = $rowsep ? $rowsep : $$tgroup{rowsep};
   $$row{valign} = $valign ? $valign : $$tgroup{$parent}{valign};
   $entryIndex = 0;
   $colspecIndex = 0;

   $result = processChildren($node);

   $rowIndex++;

   return $result;  # null expected
}


# sub entry:
# Attribute spanname is from the CALS table model, provided for backward
# compatibility only. The Exchange Model is preferred.

sub entry ($)
{
   my ($node) = @_;
   my ($result, $grandparent, $align, $char, $charoff, $colname, $colsep, 
       $morerows, $nameend, $namest, $rowsep, $spanname, $valign, $firstIndex, 
       $lastIndex, @colspecs, @spanspecs, %firstColspec, %lastColspec, 
       $tgroup, $colspec, $spanspec, $row, $entry, @keys);

   $align = $node->attribute('align');
   $char = $node->attribute('char');
   $charoff = $node->attribute('charoff');
   $colname = $node->attribute('colname');
   $colsep = $node->attribute('colsep');
   $morerows = $node->attribute('morerows');
   $nameend = $node->attribute('nameend');
   $namest = $node->attribute('namest');
   $rowsep = $node->attribute('rowsep');
   $spanname = $node->attribute('spanname');
   $valign = $node->attribute('valign');
   $grandparent = $tagStack[2];  # 'thead' | 'tbody'

   $tgroup = $table{tgroups}[$tgroupIndex];
   $row = $$tgroup{$grandparent}{rows}[$rowIndex];
   $$row{entries}[$entryIndex] = {};
   $entry = $$row{entries}[$entryIndex];
   @colspecs = @{$$tgroup{colspecs}};
   if ($$tgroup{spanspecs}) {
      @spanspecs = @{$$tgroup{spanspecs}};
   } else {
      @spanspecs = ();
   }

   # if spanname defined find spanspec
   if ($spanname) {
      foreach $spanspec (@spanspecs) {
         if ($$spanspec{spanname} eq $spanname) {
            $namest = $$spanspec{namest};
            $nameend = $$spanspec{nameend};
            last;
         }
      }
   }

   # find colspec for this entry
   $firstIndex = '';
   $lastIndex = '';
   if (!$colname && !$namest && !$nameend) {
      # use current colspec
      $firstIndex = $colspecIndex;
      $lastIndex = $colspecIndex;
      $colspecIndex++; 
   } elsif ($colname) {
      while ($colspecIndex <= $maxColspecIndex) {
         $colspec = $colspecs[$colspecIndex];
         if ($$colspec{colname} eq $colname) {
            $firstIndex = $colspecIndex;
            $lastIndex = $colspecIndex;
            $colspecIndex++; 
            last;
         } else {
            $colspecIndex++; 
         }
      }
   } else {  # $namest and $nameend should be defined
      while ($colspecIndex <= $maxColspecIndex) {
         $colspec = $colspecs[$colspecIndex];
         if ($$colspec{colname} eq $namest) {
            $firstIndex = $colspecIndex;
            $colspecIndex++; 
            last;
         } else {
            $colspecIndex++; 
         }
      }
      while ($colspecIndex <= $maxColspecIndex) {
         $colspec = $colspecs[$colspecIndex];
         if ($$colspec{colname} eq $nameend) {
            $lastIndex = $colspecIndex;
            $colspecIndex++; 
            last;
         } else {
            $colspecIndex++; 
         }
      }
   }
   if (($firstIndex eq '') || ($lastIndex eq '')) {
      print("ERROR: can't find colspec(s) for entry\n");
   }

   # ISSUE ABOUT DEFAULTS WITH ROWSEP AND PERHAPS OTHERS -
   # COLSEP OR ROW SOURCE?
   $$entry{align} = $align ? $align : $colspecs[$firstIndex]{align};
   $$entry{char} = $char ? $char : $colspecs[$firstIndex]{char};
   $$entry{charoff} = $charoff ? $charoff : $colspecs[$firstIndex]{charoff};
   $$entry{colsep} = $colsep ? $colsep : $colspecs[$firstIndex]{colsep};
   $$entry{morerows} = $morerows ? $morerows : 0;
   $$entry{rowsep} = $rowsep ? $rowsep : $$row{rowsep};
   $$entry{valign} = $valign ? $valign : $$row{valign};
   $$entry{firstIndex} = $firstIndex;
   $$entry{lastIndex} = $lastIndex;

   $result = processChildren($node);

   $grandparent = $tagStack[2];  # 'thead' | 'tbody'
   $tgroup = $table{tgroups}[$tgroupIndex];
   $row = $$tgroup{$grandparent}{rows}[$rowIndex];
   $entry = $$row{entries}[$entryIndex];
   $$entry{content} = $result;
   $entryIndex++;

   return '';
}


return 1;
