package Validate::Main;

# Module:  Validate::Main
# Purpose: Validates xml files against a DTD
# Author:  John Storrs, Laboratory for Micro Enterprise
# License: FSF GPL
# Version: $Id: Main.pm,v 1.2 2005/07/10 11:37:07 lme Exp $

# This uses the Perl regular expression engine for both element content and
# attribute content validation. In validateXml() %elementPattern and
# %attributePattern hashes are built by reference to the DTD, mapping element
# tag names to content regular expression match patterns. In validateNode,
# these patterns are tested against the actual content patterns in each xml
# file.


use strict;

BEGIN {
   require Exporter;
   our @ISA = qw(Exporter);
   our @EXPORT_OK = qw(validateXml);
}

use LME::XmlParser;
use LME::XmlNode;
use Text::Wrap qw($columns wrap);

# prototypes
sub validateXml();
sub validateNode($);

# global variables
our (%elementPattern, %attributePattern);


# sub validateXml:
# We use {} to bracket tag names in the target child pattern. See 
# validateNode() below.

sub validateXml ()
{
   my ($dtdFile, $xmlFile, $root, $result, $contents, $tag, $elementPattern, 
       $attributePattern, $attributeName, $attributeType, $attributeMode, 
       $attribute, @attributes, @contents, @xmlFiles);

   # build node tree
   $dtdFile = $ARGV[0];
   open(FILE, $dtdFile) or die("can't open $dtdFile\n");
   @contents = <FILE>;
   $contents = join('', @contents);

   # build element content pattern hash
   %elementPattern = ();
   while ($contents =~ m/<\!ELEMENT\s+(\w+)\s+([^>]+)>/g) {
                   # <!ELEMENT function (discussion?, code?)>
      $tag = $1;
      $elementPattern = $2;
      $elementPattern =~ s/(#?\w+)/({$1})/g;  # '\w' NOT GENERAL - USE [^xyz]
                   # 'section+' -> '({section})+
                   # '#PCDATA' -> '({#PCDATA})
      $elementPattern =~ s/, //g;  # strip ', '
      $elementPattern = "^$elementPattern\$";  # frame with '^', '$'
      $elementPattern{$tag} = $elementPattern;
   }

   # build attribute content pattern hash
   %attributePattern = ();
   pos($contents) = 0;
   while ($contents =~ m/<\!ATTLIST\s+(\w+)\s+([^>]+)>/g) {
                   # <!ATTLIST function {name CDATA #REQUIRED }>
                   # where the attribute pattern in braces can be repeated
      $tag = $1;
      $attributePattern = $2;
      @attributes = ();
      while ($attributePattern =~ m/([^\s]+)\s+([^\s]+)\s([^\s]+)/g) {
                                       # name CDATA #REQUIRED
         $attributeName = $1;
         $attributeType = $2;
         $attributeMode = $3;
         #print("ATTRIBUTE: $attributeName $attributeType $attributeMode\n");
         if ($attributeMode eq '#OPTIONAL') {
            $attributeName .= '?';
         }
         push(@attributes, $attributeName);
      }
      $attributePattern = join(', ', (sort(@attributes)));
      $attributePattern =~ s/([^, ?]+)/({$1})/g;
                   # 'name' -> '({name})
                   # 'label?' -> '({label})?
      $attributePattern =~ s/, //g;  # strip ', '
      $attributePattern = "^$attributePattern\$";  # frame with '^', '$'
      $attributePattern{$tag} = $attributePattern;
      #print("TAG = '$tag': PATTERN = $attributePattern\n");
   }

   close(FILE) or die("can't close $dtdFile\n");

   # validate xml files
   @xmlFiles = @ARGV[1..($#ARGV)];
   foreach $xmlFile (@xmlFiles) {
      print("validating $xmlFile\n");
      $root = parseXml(importFile("$xmlFile"));
      validateNode($root);
   }
}


# sub validateNode:
# We use '{}' to bracket tag names in the content patterns, to prevent
# erroneous matches. These characters are not legal in tag names or special
# names like #PCDATA which can occur in element patterns. This routine is
# called recursively to descend the node tree.

sub validateNode ($)
{
   my ($node) = @_;
   my ($nodeTag, $child, $childTag, $elementString, $elementPattern, 
       $attributeString, $attributePattern, $attribute, $fileName,
       $lineNumber, $message, $tagError, @attributes, @attributes2, 
       @children);

   $nodeTag = $node->tag();
   @children = $node->children();
   @attributes = $node->attributeList();
   $tagError = 0;

   # validate node element content
   $elementString = '';
   foreach $child (@children) {
      $childTag = $child->tag();
      $childTag =~ s/:text:/#PCDATA/g;
      $childTag =~ s/:cdata:/#PCDATA/g;
      $elementString .= "{$childTag}";
   }
   if ($elementString) {
      $elementPattern = $elementPattern{$nodeTag};
      if ($elementPattern) {
         if ($elementString !~ m/$elementPattern/) {
            ($fileName, $lineNumber) = $node->location();
            $message = "Element $nodeTag in file $fileName ";
            $message .= "starting at line $lineNumber ";
            $message .= "has invalid element content $elementString ";
            $message .= "which doesn't match $elementPattern.";
            $message = wrap('   ', '   ', $message);
            print("$message\n");
         }
      } else {
         print("   Don't know anything about element $nodeTag.\n");
         $tagError = 1;
      }
   }

   # validate node attribute content
   @attributes2 = ();
   foreach $attribute (@attributes) {
      push(@attributes2, ${$attribute}[0]);
   }
   $attributeString = '';
   foreach $attribute (sort(@attributes2)) {
      $attributeString .= "{$attribute}";
   }
   if ($attributeString) {
      $attributePattern = $attributePattern{$nodeTag};
      if ($attributePattern) {
         if ($attributeString !~ m/$attributePattern/) {
            ($fileName, $lineNumber) = $node->location();
            $message = "Element $nodeTag in file $fileName ";
            $message .= "starting at line $lineNumber ";
            $message .= "has invalid attribute content $attributeString ";
            $message .= "which doesn't match $attributePattern.";
            $message = wrap('   ', '   ', $message);
            print("$message\n");
         }
      } else {
         if (!$tagError) {
            print("   Don't know anything about element $nodeTag.\n");
         }
      }
   }

   # descend node tree by recursion
   foreach $child (@children) {
      validateNode($child); 
   }
}


return 1;
