package LME::XmlNode;

# Module:  LME::XmlNode
# Purpose: Provides object-oriented class for processing xml node trees.
# Author:  John Storrs, Laboratory for Micro Enterprise
# License: FSF General Public License
# Version: $Id: XmlNode.pm,v 1.2 2005/07/10 11:38:52 lme Exp $

# Provides an object-oriented class for processing xml node trees generated by 
# LME::XmlParser.


use strict;

BEGIN {
   require Exporter;
   our @ISA = qw(Exporter);
   our @EXPORT = qw($nullNode);
}

our ($nullNode);


# sub parent:
# Returns the parent of the given node.

sub parent
{
   my ($node) = @_;

   return $node->[1][0];
}


# sub firstChild:
# Returns the first child of the given node.

sub firstChild
{
   my ($node) = @_;

   return $node->[1][1];
}


# sub previousSibling:
# Returns the previous sibling of the given node.

sub previousSibling
{
   my ($node) = @_;

   return $node->[1][2];
}


# sub nextSibling:
# Returns the next sibling of the given node.

sub nextSibling
{
   my ($node) = @_;

   return $node->[1][3];
}


# sub child:
# Returns the first child of the current node satisfying the name selector, if 
# it exists.

sub child
{
   my ($node, $selector) = @_;
   my ($child, $children);

   $children = $node->[4];
   foreach $child (@$children) {
      if ($child->[0] eq $selector) {
         return $child;
      }
   }
   #print("node $selector not found - return null node\n");
   return $nullNode;
}


# sub children:
# Returns a list of all children of the current node.

sub children
{
   my ($node) = @_;
   my ($children);

   $children = $node->[4];
   if ($children) {
      return @$children;
   } else {
      return ();
   }
}


# sub tag:
# Returns the xml tag of the given node.

sub tag
{
   my ($node) = @_;
   my ($tag);

   $tag = $node->[0];

   return $tag;
}


# sub text:
# Returns the text content of the given node.
# This is a bit of a botch. It only works properly either if the node has
# text content only, or (now) if the node tag is :text:. If the node has
# mixed content, this returns the :text: children, which is probably not what 
# you want. To process mixed content, you should move through all the nodes
# recursively. Only :text: nodes really contain text.
# CLARIFY - NOT A BOTCH REALLY.

sub text
{
   my ($node) = @_;
   my ($nodeName, $child, $children);

   $nodeName = $node->[0];

   # NEW 12/9/03
   if ($nodeName  eq ':text:') {
      return $node->[2];
   }

   $children = $node->[4];
   foreach $child (@$children) {
      if ($child->[0] eq ':text:') {
         return $child->[2];
      }
   }
   #print("value not found for $nodeName\n");
   # above line causes program failure if $nodeName not defined
   return '';
}


# sub cdata:
# Returns the cdata content of the given node. See comments to text() above.

sub cdata
{
   my ($node) = @_;
   my ($nodeName, $child, $children);

   $nodeName = $node->[0];
   $children = $node->[4];

   # NEW 6/11/03
   if ($nodeName  eq ':cdata:') {
      return $node->[2];
   }

   foreach $child (@$children) {
      if ($child->[0] eq ':cdata:') {
         return $child->[2];
      }
   }
   #print("value not found for $nodeName\n");
   # above line causes program failure if $nodeName not defined
   return '';
}


# sub attribute:
# Returns the value of the named node attribute.

sub attribute
{
   my ($node, $attributeName) = @_;
   my ($nodeName, $attribute, $attributes);

   $nodeName = $node->[0];
   $attributes = $node->[3];

   foreach $attribute (@$attributes) {
      if ($attribute->[0] eq $attributeName) {
         return $attribute->[1];
      }
   }

   #print("attribute $attributeName not found for $nodeName\n");
   return '';
}


# sub setAttribute:
# Sets the value of the named node attribute.

sub setAttribute
{
   my ($node, $attributeName, $newValue) = @_;
   my ($nodeName, $attribute, $attributes);

   $nodeName = $node->[0];
   $attributes = $node->[3];

   foreach $attribute (@$attributes) {
      if ($attribute->[0] eq $attributeName) {
         $attribute->[1] = $newValue;
         return;
      }
   }
   #print("attribute $attributeName not found for $nodeName\n");
}


# sub location:
# Returns the file name and line number where the given node starts.

sub location
{
   my ($node) = @_;
   my ($fileName, $charNumber, $lineNumber, $xml, @contents);

   $fileName = $node->[5][0];
   $charNumber = $node->[5][1];

   open(FILE, $fileName) or die("can't open $fileName\n");
   @contents = <FILE>;
   $xml = join('', @contents);
   $lineNumber = 1;
   while ($xml =~ m/\G.*\n/g) {
      $lineNumber++;
      if (pos($xml) > $charNumber) {
         $lineNumber--;
         last;
      }
   }

   return ($fileName, $lineNumber);
}


# sub nodeList:
# Function interface to nodelistRecursive() below, returning an array 
# of nodes. Parameter depth is optional. If it is defined it limits the
# search depth. For example, depth = 1 will return children of the current
# node only.

sub nodeList
{
   my ($node, $selector, $depth) = @_;
   my (@nodeList);

   @nodeList = ();

   if (!defined($depth)) {
      $depth = 1e6;  # unlimited
   }
   nodeListRecursive($node, $selector, \@nodeList, $depth);

   return @nodeList;
}


# sub attributeList:
# Returns the node attribute list.

sub attributeList
{
   my ($node) = @_;
   my ($attributes);

   $attributes = '';
   if (($node->tag() ne ':text:') &&
       ($node->tag() ne ':cdata:')) {
   $attributes = $node->[3];
   }

   if ($attributes) {
      return @{$attributes};
   } else {
      return ();
   }
}


# sub generateXml:
# Function interface to generateXmlRecursive() below, returning an xml string.

sub generateXml
{
   my ($node) = @_;
   my ($xml, $indentLevel, $indentString);

   $xml = '';
   $indentLevel = 0;  # could be a global to allow configuration
   $indentString = '  ';  # could be a global to allow configuration

   generateXmlRecursive($node, $indentLevel, $indentString, \$xml);

   return $xml;
}


# sub generateTable:
# Generates a DocBook table representation of the given node. If the title 
# parameter is not an empty string, the table is formal, otherwise informal.

sub generateTable ($$$$)
{
   my ($node, $indentLevel, $indentString, $title) = @_;
   my ($name, $attribute, $attributes, $children, $child, $indent, $result);

   $name = $node->[0];
   $attributes = $node->[3];
   $children = $node->[4];
   $result = '';
   
   $indent = $indentString x $indentLevel;
   if ($title ne '') {
      $result .= "$indent<table>\n";
      $indent = $indentString x ++$indentLevel;
      $result .= "$indent<title>$title</title>\n";
   } else {
      $result .= "$indent<informaltable>\n";
      $indent = $indentString x ++$indentLevel;
   }

   $result .= "$indent<tgroup cols=\"2\">\n";

   $indent = $indentString x ++$indentLevel;
   $result .= "$indent<colspec colnum=1 colwidth=\"4.0cm\">\n";   
   $result .= "$indent<colspec colnum=2 colwidth=\"12.0cm\">\n";   
   #$result .= "$indent<thead>\n";
   #$result .= "$indent</thead>\n";

   $result .= "$indent<tbody>\n";
   $indent = $indentString x ++$indentLevel;

   foreach $attribute (@$attributes) {
      $result .= "$indent<row>\n";

      $indent = $indentString x ++$indentLevel;
      $result .= "$indent<entry>";
      $result .= "\@$attribute->[0]";
      $result .= "</entry>\n";
      $result .= "$indent<entry>";
      $result .= "$attribute->[1]";
      $result .= "</entry>\n";

      $indent = $indentString x --$indentLevel;
      $result .= "$indent</row>\n";
   }

   foreach $child (@$children) {
      $result .= "$indent<row>\n";

      $indent = $indentString x ++$indentLevel;
      $result .= "$indent<entry>";
      $result .= "$child->[0]";
      $result .= "</entry>\n";
      $result .= "$indent<entry>";
      $result .= nodeValue($child);
      $result .= "</entry>\n";

      $indent = $indentString x --$indentLevel;
      $result .= "$indent</row>\n";
   }
   $indent = $indentString x --$indentLevel;
   $result .= "$indent</tbody>\n";

   $indent = $indentString x --$indentLevel;
   $result .= "$indent</tgroup>\n";

   $indent = $indentString x --$indentLevel;
   if ($title ne '') {
      $result .= "$indent</table>\n";
   } else {
      $result .= "$indent</informaltable>\n";
   }

   return $result;
}


# sub generateVariableList:
# Generates a DocBook variable list representation of the given node. If the 
# title parameter is not an empty string, the list has a title.

sub generateVariableList ($$$$)
{
   my ($node, $indentLevel, $indentString, $title) = @_;
   my ($name, $attribute, $attributes, $children, $child, $indent, $value, 
       $result);

   $name = $node->[0];
   $attributes = $node->[3];
   $children = $node->[4];
   $result = '';
   
   $indent = $indentString x $indentLevel;
   $result .= "$indent<variablelist>\n";

   $indent = $indentString x ++$indentLevel;
   if ($title ne '') {
      $result .= "$indent<title>$title</title>\n";
   }

   foreach $attribute (@$attributes) {
      $result .= "$indent<varlistentry>\n";

      $indent = $indentString x ++$indentLevel;
      $result .= "$indent<term><emphasis>\@$attribute->[0]</emphasis></term>\n";
      $result .= "$indent<listitem><para>$attribute->[1]</para></listitem>\n";

      $indent = $indentString x --$indentLevel;
      $result .= "$indent</varlistentry>\n";
   }

   foreach $child (@$children) {
      $result .= "$indent<varlistentry>\n";

      $value = nodeValue($child);
      $indent = $indentString x ++$indentLevel;
      $result .= "$indent<term><emphasis>$child->[0]</emphasis></term>\n";
      $result .= "$indent<listitem><para>$value</para></listitem>\n";

      $indent = $indentString x --$indentLevel;
      $result .= "$indent</varlistentry>\n";
   }
   $indent = $indentString x --$indentLevel;
   $result .= "$indent</variablelist>\n";

   return $result;
}


# sub destroy:
# Implements garbage collection.

sub destroy
{
   my ($node) = @_;
   my ($name, $links, $attribute, $attributes, $child, $children, $link);

   $name = $node->[0];
   $links = $node->[1];
   $attributes = $node->[3];
   $children = $node->[4];

   undef($name);

   foreach $link (@$links) {
      undef($link);
   }
   
   foreach $attribute (@$attributes) {
      undef($attribute->[0]);  # attribute name
      undef($attribute->[1]);  # attribute value
   }

   foreach $child (@$children) {
      if ($child->[0] eq ':text:') {
         undef($child->[2]);  # text content
      }
      destroy($child);
   }
   undef(@$links);
   undef(@$attributes);
   undef(@$children);
   undef($links);
   undef($attributes);
   undef($children);
   undef($node);
}


# PRIVATE

# sub nodeListRecursive:
# A recursive selector subroutine which returns the $nodeList reference 
# parameter. Selects nodes with the given node name in standard tree order,
# limiting the search to the given depth. Call through nodeList() above. 

sub nodeListRecursive
{
   my ($node, $selector, $nodeList, $depth) = @_;
   my ($child, $children);

   $children = $node->[4];
   foreach $child (@$children) {
      if ($child->[0] eq $selector) {
         push(@$nodeList, $child);
      }
      $depth--;
      if ($depth > 0) {
         nodeListRecursive($child, $selector, $nodeList, $depth);
      }
   }
}


# sub generateXmlRecursive:
# Recursively generates xml for the given node, returning it in the
# $xml reference parameter. Call through generateXml() above.

sub generateXmlRecursive
{
   my ($node, $indentLevel, $indentString, $xml) = @_;
   my ($name, $attribute, $attributes, $child, $children, $indent, $hasCdata);

   $name = $node->[0];
   $attributes = $node->[3];
   $children = $node->[4];
   $indent = $indentString x $indentLevel;
   $hasCdata = 0;
   
   $$xml .= "\n$indent<$name";
   foreach $attribute (@$attributes) {
      $$xml .= " $attribute->[0]=\"$attribute->[1]\"";
   }
   #if (@$children) {  
   if ($children && @$children) {  # TEMPORARY FIX FOR EMPTY PARAS ETC
      $$xml .= ">";
   } else {
      $$xml .= "/>";
   }

   foreach $child (@$children) {
      if ($child->[0] eq ':text:') {
         $$xml .= "$child->[2]";  # value
         $hasCdata = 1;
      } elsif ($child->[0] eq ':cdata:') {
         $$xml .= "<![CDATA[\n";
         $$xml .= "$child->[2]";  # value
         $$xml .= "]]>\n";
         $hasCdata = 1;
      }
      else {
         generateXmlRecursive($child, $indentLevel + 1, $indentString, $xml);
      }
   }

   if (@$children) {
      if ($hasCdata != 1) {
         $$xml .= "\n$indent";
      }
      $$xml .= "</$name>";
   }
}


# Implements an abbreviated syntax for child node processing.
# DOESN'T WORK YET. When this is enabled, get error messages on program
# termination because this is called to provide DESTROY!

=hide
sub AUTOLOAD 
{
   my ($node) = @_;
   my ($element) = $AUTOLOAD;
   my ($listContext);

   $element =~ s/^.*:://;  # strip package name

   $listContext = wantarray();  # reports subroutine context
   if ($listContext) {
      return $node->nodeList($element);
   } else {
      return $node->child($element)->text();
   }
}
=cut


$nullNode->[0] = 'null';

INIT {
   bless $nullNode;
}


# Overloading:
# Perl object overloading is rather fraught with complications. Here we just 
# want to define bool, but this requires that we also define "" and 0+ as well, 
# plus provide a fallback. Took a day to suss this out, not helped by the lack 
# of discussion about it.

use overload 
   'bool' => \&boolify,
   '""' => \&stringify,
   '0+' => \&numify,
   'fallback' => 1;  # allows default ==, != etc to be used

sub boolify
{
   my ($node) = @_;

   return (\@{$node} != \@{$nullNode});
}

sub stringify
{
   my ($node) = @_;

   return \@{$node};
}

sub numify
{
   my ($node) = @_;

   return \@{$node};
}


return 1;
