package LME::XmlParser;

# Module:  LME::XmlParser
# Purpose: Provides an incremental  parser system for xml files.
# Author:  John Storrs, Laboratory for Micro Enterprise
# License: FSF General Public License
# Version: $Id: XmlParser.pm,v 1.2 2005/07/10 11:38:52 lme Exp $

# We use an array-based structure to store XML node (element) data. XML nodes 
# are references (pointers) to these structures:
#    $node->[0]: tag name (string);
#    $node->[1][0]: parent (node reference);
#    $node->[1][1]: first child (node reference);
#    $node->[1][2]: previous sibling (node reference);
#    $node->[1][3]: next sibling (node reference);
#    text element (tag :text:):
#       $node->[2]: text (string);
#    cdata element (tag :cdata:):
#       $node->[2]: cdata text (string);
#    comment element (tag :comment:):
#       $node->[2]: comment text (string);
#    doctype element (tag :doctype:):
#       $node->[2]: doctype text (string);
#    $node->[3][n][0]: name of attribute n (string);
#    $node->[3][n][1]: value of attribute n (string);
#    $node->[4]: children (node reference array);
#    $node->[5][0]: source file name (string);
#    $node->[5][1]: element start index (integer).
# Special tag names are used as indicated to identify text, cdata, comment and
# doctype elements. Normal element attribute indices (n above) range from 0
# to attribute count - 1. The last two values are used for error reporting; the
# element start index locates the opening '<'. Nodes are exported by the parser
# as Perl objects. Users of the XmlNode object-oriented API are unaware of the
# structure described above. They view the parsed XML as a node tree, navigated
# and processed using the API.

# We use arrays for data organisation to optimise memory/speed performance.
# Initial tests using hashes for attributes showed no significant speed 
# increase. In any case, attribute order should be preserved when XML is to be 
# generated from the node tree, so value hashes alone are insufficient. Array
# searches are made for both named elements and named attributes. These don't
# seem to affect performance significantly. A possible further optimisation
# would be to add element and attribute fast access hashes to the present node 
# structure.

# BUGS:
# Recursive includes don't maintain correct cwd on return. That is, if
# an included file includes others which change the include path, the old path
# isn't restored when that include file exits. Needs a stack, and the knowledge
# of when to pop it.


use strict;

BEGIN {
   require Exporter;
   our @ISA = qw(Exporter);
   our @EXPORT = 
      qw(importFile importFile2 nextNode nextChild parseXml startParse);
}

use LME::XmlNode;

sub importFile($);
sub importFile2($);
sub nextNode();
sub nextChild($);
sub parseXml($);
sub startParse($);
sub lineNumber($$);

our ($xml, $finish, $nestingLevel, $nodeBuffer, @nodeStack, $nullNode, 
     $filename, $endOfXml, $includedElement, $importedFile, 
     $ignoreDoctype, $ignoreComment, $ignoreInstruction, $ignoreWhiteSpace, 
     $includePath, @includeStack);


INIT {
   $ignoreDoctype = 1;
   $ignoreComment = 1;
   $ignoreInstruction = 1;
   $ignoreWhiteSpace = 1;
   $nullNode = $LME::XmlNode::nullNode;
   $nodeBuffer = $nullNode;
   $nodeStack[0] = $nullNode;
   @includeStack = ();
}

# sub importFile:
# Imports a file and returns a string containing the file contents. Treats 
# filenames as relative to the current working directory of the application. 
# Resets the include path, so inclusion using importFile2() handles relative
# pathnames correctly.

sub importFile ($)
{
   my ($pathname) = @_;
   my ($contents, @contents, $directory);

   $pathname =~ s/^\s+//;  # strip enclosing white space
   $pathname =~ s/\s+$//;
   if ($pathname =~ m|/|) {
      $includePath = $pathname;
      $includePath =~ s|([^/]*)$||;  # strip basename
   } else { 
      $includePath = '';
   }

   $filename = $pathname;
   if (-f $filename) {
      open(FILE, $filename) or die("can't open $filename\n");
      @contents = <FILE>;
      $contents = join('', @contents);
      $importedFile = 1;
      close(FILE) or die("can't close $filename\n");
   } else {
      die("can't find $filename\n");
   }

   return $contents;
}


# sub importFile2:
# Imports a file into global $xml and initializes control variables. Keeps 
# track of the current directory and knows about relative and absolute 
# pathnames.

# Can we combine this with importFile? If not, rename as includeFile?

sub importFile2 ($)
{
   my ($pathname) = @_;
   my ($contents, @contents, $basename, $directory);

   if ($pathname =~ m|/|) {
      $directory = $pathname;
      $directory =~ s|([^/]*)$||;  # strip basename
      $basename = $1;
      if ($directory =~ m|^/|) {  # pathname is absolute
         $includePath = $directory;  # replace
      } else {
         $includePath .= $directory;  # append
      }
   } else {
      $basename = $pathname;
   }

   $filename = "${includePath}$basename";
   if (-f $filename) {
      open(FILE, $filename) or die("can't open $filename\n");
      @contents = <FILE>;
      $xml = join('', @contents);
      close(FILE) or die("can't close $filename\n");
   } else {
      die("can't find $filename\n");
   }

   $finish = 0;
   $nestingLevel = 0;
   $includedElement = 0;
   $endOfXml = 0;
}


# sub nextNode ():
# Returns the next node in standard tree order if there is one in the input
# stream. Otherwise returns the undefined value and sets the global variable
# $endOfXml. The input stream may contain include files specified using the
# special <include> tag. This may have element and id attributes to identify a
# specific element for inclusion. Node types are processed in order of likely 
# frequency for efficiency.

sub nextNode ()
{
   my ($attributes, $attributeCount, $cdata, $closingTag, $comment, $emptyTag,
       $fullTag, $indent, $indentString, $index, $openingTag, $tagLength, 
       $tagName, $text, $textLength, $parent, $node, @attribute, $doctype,
       @attributeValue, $start, $stackNode, $stackTag, $tagFile, $tagStart,
       $lineNumber, $includeFile, $requiredElement, $requiredId);

   # get next wanted node
   $node = undef;
   do {
      pos($xml) = $finish;
      $parent = $nodeStack[0];
   
      if ($xml =~ m{  # process text node
                           \G       # force immediate match
                           ([^<]+)  # $1 = any text up to but not including '<'
                        }gcx) {
         $text = $1;
         $textLength = length($text);
         $finish += $textLength;
         if (!($ignoreWhiteSpace && ($text =~ m/^\s+$/))) {
            $node->[0] = ':text:';
            $node->[1][0] = $parent;
            $node->[2] = $text;
         }
   
      } elsif ($xml =~ m{  # process element node
                           (\G             # $1 force immediate match
                           <               # opening '<'
                           [\s\n]*         # optional whitespace
                           (/?)            # $2 = optional '/'
                           [\s\n]*         # optional whitespace
                           ([\w\-:]+)      # $3 = tag name
                           [\s\n]*         # optional whitespace
                           (([^>]*['"])?)  # $4 = attributes
                           [\s\n]*         # optional whitespace
                           (/?)            # $6 = optional '/'
                           [\s\n]*         # optional whitespace
                           >)              # closing '>'
                        }gcx) {
         $fullTag = $1;
         $closingTag = ($2 eq '/');
         $tagName = $3;
         $attributes = $4;
         $emptyTag = ($6 eq '/');
         $tagLength = length($fullTag);
         $start = $finish;
         $finish += $tagLength;
         $openingTag = (!$closingTag) && (!$emptyTag);
         if ($closingTag && $emptyTag) {
            warn("opening and closing tag symbols both set\n");
         }
   
         # extract attributes into arrays, maintaining ordering
         @attribute = ();
         @attributeValue = ();
         $index = 0;
         while ($attributes =~ m{
                                   ([\w\-]+)      # $1 = attribute name
                                   \s*=\s*        # ' = '
                                   (["'])(.*?)\2  # $3 = attribute value
                                }gcsx) {
            $attribute[$index] = $1;
            $attributeValue[$index] = $3;
            $index++;
         }
         $attributeCount = $index;
   
         # construct element node
         if ($openingTag || $emptyTag) { 
            $node->[0] = $tagName;
            $node->[1][0] = $parent;
            foreach $index (0 .. $attributeCount - 1) {
               $node->[3][$index][0] = $attribute[$index];
               $node->[3][$index][1] = $attributeValue[$index];
            }
            $node->[5][0] = $filename;  # for error reporting
            $node->[5][1] = $start;  # ditto
         }
   
      } elsif ($xml =~ m{  # process CDATA node
                         (\G          # force immediate match
                         <!\[CDATA\[  # opening '<![CDATA[' 
                         (.*?)        # $2 = CDATA text, minimal match
                         \]\]>)       # closing ']]>'
                      }gcsx) {
         $fullTag = $1;
         $cdata = $2;
         $tagLength = length($fullTag);
         $finish += $tagLength;
         $node->[0] = ':cdata:';
         $node->[1][0] = $parent;
         $node->[2] = $cdata;
   
      } elsif ($xml =~ m{  # process comment node
                           (\G    # force immediate match
                           <!--   # opening '<!--' 
                           (.*?)  # $2 = comment text, minimal match
                           -->)   # closing '-->'
                        }gcsx) {
         $fullTag = $1;
         $comment = $2;
         $tagLength = length($fullTag);
         $finish += $tagLength;
         if (!$ignoreComment) {
            $node->[0] = ':comment:';
            $node->[1][0] = $parent;
            $node->[2] = $comment;
         }
   
      } elsif ($xml =~ m{  # process doctype node
                           (\G        # force immediate match
                           <!DOCTYPE  # opening '<!--' 
                           (.*?)      # $2 = comment text, minimal match
                           >)         # closing '>'
                        }gcsx) {
         $fullTag = $1;
         $doctype = $2;
         $tagLength = length($fullTag);
         $finish += $tagLength;
         if (!$ignoreDoctype) {
            $node->[0] = ':doctype:';
            $node->[1][0] = $parent;
            $node->[2] = $doctype;
         }
   
      } else {  # end of file or match failure
         if (!$endOfXml && ($nestingLevel != 0)) {
            pop(@nodeStack);  # remove $nullNode in last position
            print("nesting error - unterminated elements:\n");
            foreach $stackNode (@nodeStack) {
               $stackTag = $stackNode->[0];
               $tagFile = $stackNode->[5][0];
               $tagStart = $stackNode->[5][1];
               $lineNumber = lineNumber($tagFile, $tagStart);
               print("   tag '$stackTag' starting at ");
               print("$tagFile:$lineNumber\n");
            }
         }
         $endOfXml = 1;
      }

      # make $node a Perl object before copy operations
      if ($node) {
         bless($node, "LME::XmlNode");
      }
   
      # manage node stack and report nesting errors
      if ($openingTag) { 
         unshift(@nodeStack, $node);  # push onto stack
         $nestingLevel++;
      } elsif ($closingTag) { 
         $stackNode = shift(@nodeStack);  # pop from stack
         $nestingLevel--;
         if ($stackNode->tag() ne $tagName) {  # nesting error
            $stackTag = $stackNode->[0];
            $tagFile = $stackNode->[5][0];
            $tagStart = $stackNode->[5][1];
            print("nesting error: \n");
            $lineNumber = lineNumber($tagFile, $tagStart);
            print("   closing tag for '$stackTag' starting at ");
            print("$tagFile:$lineNumber expected,\n");
            $lineNumber = lineNumber($filename, $start);
            print("   but closing tag for '$tagName' found, ");
            print("at $filename:$lineNumber\n\n");
            $stackNode = shift(@nodeStack);  # pop from stack
            $nestingLevel--;
         }
      }

      # terminate include file if required element found
      if ($includedElement && 
          ($closingTag || $emptyTag) && ($nestingLevel == 0)) {
         $endOfXml = 1;
         $node = undef;
      }

      # restore previous file if include file terminated
      if ($endOfXml && @includeStack) {
         $filename = $includeStack[0]{filename};
         $xml = $includeStack[0]{xml};
         $finish = $includeStack[0]{finish};
         $nestingLevel = $includeStack[0]{nestingLevel};
         $endOfXml = $includeStack[0]{terminated};
         $includedElement = $includeStack[0]{includedElement};
         @nodeStack = @{$includeStack[0]{nodeStack}};
         shift(@includeStack);  # pop from stack
      }

      # handle file inclusion
      if ($node && ($node->[0] eq 'include')) {
         $includeFile = $node->attribute('file');     
         $requiredElement = $node->attribute('element');     
         $requiredId = $node->attribute('id');     
         $node = undef;  # force read of next node from new file

         # treatment depends on whether include element is empty
         # or contains nested elements which need processing
         # for now just do this - empty element expected
         if ($openingTag) {
            shift(@nodeStack);  # pop include node from stack
            $nestingLevel--;
         }

         # save current file state
         unshift(@includeStack, {});  # push onto stack
         $includeStack[0]{filename} = $filename;
         $includeStack[0]{xml} = $xml;
         $includeStack[0]{finish} = $finish;
         $includeStack[0]{nestingLevel} = $nestingLevel;
         $includeStack[0]{terminated} = $endOfXml;
         $includeStack[0]{includedElement} = $includedElement;
         $includeStack[0]{nodeStack} = [@nodeStack];

         # import new include file
         importFile2($includeFile);  # puts file contents in $xml
      }

      # seek to required element if specified
      if ($requiredElement && !$endOfXml) {
         if ($node && ($node->[0] eq $requiredElement) && 
               ($node->attribute('id') eq $requiredId)) {
            @nodeStack = @{$includeStack[0]{nodeStack}};
            $node->[1][0] = $nodeStack[0];  # fix parent
            $nestingLevel = 1;  # fix nesting level
            $includedElement = 1;
            if ($openingTag) {
               unshift(@nodeStack, $node);  # push onto stack
            }
         } else {
            $node = undef;  # continue seek
         }
      }
      $openingTag = 0;
      $closingTag = 0;
      $emptyTag = 0;

   } until ($node || $endOfXml);

   return $node;
}


sub nextChild ($)
{
   my ($parent) = @_;
   my ($bufferCopy);

   if (!$nodeBuffer) {
      $nodeBuffer = nextNode();
   }

   # return child node to application
   if (!$endOfXml && ($nodeBuffer->[1][0] == $parent)) {
      $bufferCopy = $nodeBuffer;
      $nodeBuffer = undef;
      return $bufferCopy;
   } else {
      return undef;
   }
}


# sub parseXml:
# Builds an in-memory tree representation of the xml string, for processing by 
# the LME::XmlNode library. The nodes are provided by calls to nextNode(). This 
# routine binds the nodes into a redundantly linked tree for efficient 
# processing.

sub parseXml ($)
{
   my ($xmlString) = @_;
   my ($rootNode, $currentNode, $parentNode);

   if ($importedFile) {
      $importedFile = 0;
   } else {
      $filename = 'string';
   }
   $xml = $xmlString;
   $finish = 0;
   $nestingLevel = 0;
   $includedElement = 0;
   $endOfXml = 0;

   $rootNode = nextNode();
   while (!$endOfXml) {
      $currentNode = nextNode();
      $parentNode = $currentNode->[1][0]; 
      push(@{$parentNode->[4]}, $currentNode);  # add child
      if (scalar(@{$parentNode->[4]}) == 1) {  # only 1 child
         $parentNode->[1][1] = $currentNode;  # first child
         $currentNode->[1][2] = $nullNode;  # previous sibling
      } else {  # more than 1 child
         $currentNode->[1][2] = ${$parentNode->[4]}[-2];  # previous sibling
         ${$parentNode->[4]}[-2]->[1][3] = $currentNode;  # next sibling
      }
      $currentNode->[1][3] = $nullNode;  # next sibling;
   }

   return $rootNode;
}


# sub startParse:
# Starts an incremental parse of the xml string parameter.

sub startParse ($)
{
   my ($xmlString) = @_;

   if ($importedFile) {
      $importedFile = 0;
   } else {
      $filename = 'string';
   }
   $xml = $xmlString;
   $finish = 0;
   $nestingLevel = 0;
   $includedElement = 0;
   $endOfXml = 0;
}


sub lineNumber ($$)
{
   my ($filename, $charNumber) = @_;
   my (@contents, $xml, $line);

   open(FILE, $filename) or die("can't open $filename\n");
   @contents = <FILE>;
   $xml = join('', @contents);
   $line = 1;
   while ($xml =~ m/\G.*\n/g) {
      $line++;
      if (pos($xml) > $charNumber) {
         $line--;
         last;
      }
   }
   return $line;
}


return 1;
